# populator/helpers/luks.py
# LUKS backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from ... import udev
from ...devices import LUKSDevice, IntegrityDevice
from ...errors import DeviceError, LUKSError
from ...flags import flags
from .devicepopulator import DevicePopulator
from .formatpopulator import FormatPopulator
from ...static_data import luks_data, stratis_info

import logging
log = logging.getLogger("blivet")


class LUKSDevicePopulator(DevicePopulator):
    @classmethod
    def match(cls, data):
        return udev.device_is_dm_luks(data)

    def run(self):
        parents = self._devicetree._add_parent_devices(self.data)
        device = LUKSDevice(udev.device_get_name(self.data),
                            sysfs_path=udev.device_get_sysfs_path(self.data),
                            parents=parents,
                            exists=True)
        self._devicetree._add_device(device)
        return device


class IntegrityDevicePopulator(DevicePopulator):
    @classmethod
    def match(cls, data):
        return udev.device_is_dm_integrity(data)

    def run(self):
        parents = self._devicetree._add_parent_devices(self.data)
        name = udev.device_get_name(self.data)

        try:
            info = blockdev.crypto.integrity_info(name)
            # integrity algorithm is not part of the on-disk metadata but for active
            # device we can get it from device mapper so let's set it here
            parents[0].format.algorithm = info.algorithm
        except blockdev.BlockDevError:
            log.info("failed to get information about integrity device %s", name)

        device = IntegrityDevice(name,
                                 sysfs_path=udev.device_get_sysfs_path(self.data),
                                 parents=parents,
                                 exists=True)
        self._devicetree._add_device(device)
        return device


class LUKSFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "luks"

    @classmethod
    def match(cls, data, device):  # pylint: disable=arguments-differ,unused-argument
        if not super(LUKSFormatPopulator, cls).match(data, device):
            return False

        # locked stratis pools are managed in the StratisFormatPopulator
        for pool in stratis_info.locked_pools:
            if device.path in pool.devices:
                return False

        # unlocked stratis pools are also managed in the StratisFormatPopulator
        holders = udev.device_get_holders(data)
        if holders:
            fs = udev.device_get_format(holders[0])
            if fs == "stratis":
                log.debug("ignoring LUKS format on %s, it appears to be an encrypted Stratis pool",
                          device.name)
                return False

        return True

    def _get_kwargs(self):
        kwargs = super(LUKSFormatPopulator, self)._get_kwargs()

        holders = udev.device_get_holders(self.data)
        if holders:
            kwargs["name"] = udev.device_get_name(holders[0])
        else:
            kwargs["name"] = "luks-%s" % udev.device_get_uuid(self.data)

        kwargs["luks_version"] = "luks%s" % udev.device_get_format_version(self.data)
        return kwargs

    def run(self):
        super(LUKSFormatPopulator, self).run()
        if not self.device.format.uuid:
            log.info("luks device %s has no uuid", self.device.path)
            return

        # look up or create the mapped device
        if not self._devicetree.get_device_by_name(self.device.format.map_name):
            passphrase = luks_data.luks_devs.get(self.device.format.uuid)
            if self.device.format.configured:
                pass
            elif passphrase:
                self.device.format.passphrase = passphrase
            elif self.device.format.uuid in luks_data.luks_devs:
                log.info("skipping previously-skipped luks device %s",
                         self.device.name)
            elif self._devicetree._cleanup or flags.testing:
                # if we're only building the devicetree so that we can
                # tear down all of the devices we don't need a passphrase
                if self.device.format.status:
                    # this makes device.configured return True
                    self.device.format.passphrase = 'yabbadabbadoo'
            else:
                # Try each known passphrase. Include luks_data.luks_devs values in case a
                # passphrase has been set for a specific device without a full
                # reset/populate, in which case the new passphrase would not be
                # in luks_data.passphrases.
                passphrases = luks_data.passphrases + list(luks_data.luks_devs.values())
                for passphrase in passphrases:
                    self.device.format.passphrase = passphrase
                    try:
                        self.device.format.setup()
                    except blockdev.BlockDevError:
                        self.device.format.passphrase = None
                    else:
                        break

            # try only to setup the luks format -- the luks device will be
            # discovered and added later by the LUKSDevicePopulator
            try:
                self.device.format.setup()
            except (LUKSError, blockdev.CryptoError, DeviceError) as e:
                log.info("setup of %s failed: %s", self.device.format.map_name, e)
        else:
            log.warning("luks device %s already in the tree",
                        self.device.format.map_name)


class IntegrityFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "integrity"

    def _get_kwargs(self):
        kwargs = super(IntegrityFormatPopulator, self)._get_kwargs()

        holders = udev.device_get_holders(self.data)
        if holders:
            kwargs["name"] = udev.device_get_name(holders[0])
        else:
            # this is just a fallback for closed and non-existing integrity devices
            # similar to LUKS devices where we assume the name of the active device
            # is luks-<UUID> (integrity format has no UUID so we are using name here)
            kwargs["name"] = "integrity-%s" % udev.device_get_name(self.data)
        return kwargs



#
# nvme.py - NVMe class
#
# Copyright (C) 2022  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil

from . import errors
from . import util

import logging
log = logging.getLogger("blivet")

HOSTNQN_FILE = "/etc/nvme/hostnqn"
HOSTID_FILE = "/etc/nvme/hostid"


class NVMe(object):
    """ NVMe utility class.

        .. warning::
            Since this is a singleton class, calling deepcopy() on the instance
            just returns ``self`` with no copy being created.
    """

    def __init__(self):
        self.started = False

    # So that users can write nvme() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):  # pylint: disable=unused-argument
        return self

    def startup(self):
        if self.started:
            return

        rc, nqn = util.run_program_and_capture_output(["nvme", "gen-hostnqn"])
        if rc != 0:
            raise errors.NVMeError("Failed to generate hostnqn")

        with open(HOSTNQN_FILE, "w") as f:
            f.write(nqn)

        rc, hid = util.run_program_and_capture_output(["dmidecode", "-s", "system-uuid"])
        if rc != 0:
            raise errors.NVMeError("Failed to generate host ID")

        with open(HOSTID_FILE, "w") as f:
            f.write(hid)

        self.started = True

    def write(self, root):  # pylint: disable=unused-argument
        # copy the hostnqn and hostid files
        if not os.path.isdir(root + "/etc/nvme"):
            os.makedirs(root + "/etc/nvme", 0o755)
        shutil.copyfile(HOSTNQN_FILE, root + HOSTNQN_FILE)
        shutil.copyfile(HOSTID_FILE, root + HOSTID_FILE)


# Create nvme singleton
nvme = NVMe()
""" An instance of :class:`NVMe` """



import abc

from six import add_metaclass


@add_metaclass(abc.ABCMeta)
class FSUUID(object):

    """An abstract class that represents filesystem actions for setting the
       UUID.
    """

    @classmethod
    @abc.abstractmethod
    def uuid_format_ok(cls, uuid):
        """Returns True if the given UUID is correctly formatted for
           this filesystem, otherwise False.

           :param str uuid: the UUID for this filesystem
           :rtype: bool
        """
        raise NotImplementedError

    # IMPLEMENTATION methods

    @classmethod
    def _check_rfc4122_uuid(cls, uuid):
        """Check whether the given UUID is correct according to RFC 4122 and
           return True if it's correct or False otherwise.

           :param str uuid: the UUID to check
           :rtype: bool
        """
        chunks = uuid.split('-')
        if len(chunks) != 5:
            return False
        chunklens = [len(chunk) for chunk in chunks
                     if all(char in "0123456789abcdef" for char in chunk)]
        return chunklens == [8, 4, 4, 4, 12]


class Ext2FSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        return cls._check_rfc4122_uuid(uuid)


class FATFSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        if len(uuid) != 9 or uuid[4] != '-':
            return False
        return all(char in "0123456789ABCDEF"
                   for char in (uuid[:4] + uuid[5:]))


class JFSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        return cls._check_rfc4122_uuid(uuid)


class ReiserFSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        return cls._check_rfc4122_uuid(uuid)


class XFSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        return cls._check_rfc4122_uuid(uuid)


class HFSPlusUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        return cls._check_rfc4122_uuid(uuid)


class NTFSUUID(FSUUID):
    @classmethod
    def uuid_format_ok(cls, uuid):
        if len(uuid) != 16:
            return False
        return all(char in "0123456789ABCDEF" for char in uuid)



#
# Copyright (C) 2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import dbus
import dbus.service

from .constants import BUS_NAME


class DBusObject(dbus.service.Object):
    """ Base class for dbus objects. """
    def __init__(self, manager):
        # pylint: disable=super-init-not-called
        self._present = True
        self._init_dbus_object()
        self._manager = manager  # provides ObjectManager interface

    # This is here to make it easier to prevent the dbus.service.Object
    # constructor from running during unit testing.
    def _init_dbus_object(self):
        """ Initialize superclass. """
        super(DBusObject, self).__init__(bus_name=dbus.service.BusName(BUS_NAME, dbus.SystemBus()),
                                         object_path=self.object_path)

    @property
    def present(self):
        """ Is this object present in blivet's current view? """
        return self._present

    @present.setter
    def present(self, state):
        """ Indicate whether the object is in blivet's current view. """
        self._present = state

    def remove_from_connection(self, connection=None, path=None):
        super(DBusObject, self).remove_from_connection(connection=connection, path=path)
        self._object_path = None

    @property
    def id(self):
        """ The unique id of this instance. """
        raise NotImplementedError()

    @property
    def object_path(self):
        """ The dbus object path for this instance. """
        raise NotImplementedError()

    @property
    def interface(self):
        """ The interface implemented by this class. """
        raise NotImplementedError()

    @property
    def properties(self):
        """ dict of property key/value pairs to export via dbus. """
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface_name):
        if interface_name != self.interface:
            raise dbus.exceptions.DBusException('%s.UnknownInterface' % BUS_NAME,
                                                'The %s object does not implement the %s interface'
                                                % (self.__class__.__name__, interface_name))

        return self.properties

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature='ss', out_signature='v')
    def Get(self, interface_name, property_name):
        return self.GetAll(interface_name)[property_name]

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature='ssv')
    def Set(self, interface_name, property_name, new_value):
        if interface_name != self.interface:
            raise dbus.exceptions.DBusException('%s.UnknownInterface' % BUS_NAME,
                                                'The %s object does not implement the %s interface'
                                                % (self.__class__.__name__, interface_name))

        if property_name in self.properties:
            raise dbus.exceptions.DBusException('%s.ReadOnlyProperty' % BUS_NAME,
                                                'The %s property is read-only' % property_name)
        else:
            raise dbus.exceptions.DBusException('%s.UnknownProperty' % BUS_NAME,
                                                'The %s interface does not have %s property'
                                                % (interface_name, property_name))

        self.PropertiesChanged(interface_name, {property_name: new_value}, [])

    @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE, signature='sa{sv}as')
    def PropertiesChanged(self, interface_name, changed_properties, invalidated_properties):
        pass



# dmraid.py
# dmraid device formats
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

from ..storage_log import log_method_call
from ..errors import DMRaidMemberError
from ..i18n import N_
from . import DeviceFormat, register_device_format

import logging
log = logging.getLogger("blivet")


class DMRaidMember(DeviceFormat):

    """ A dmraid member disk. """
    _type = "dmraidmember"
    _name = N_("dm-raid member device")

    _udev_types = ["adaptec_raid_member", "hpt37x_raid_member", "hpt45x_raid_member",
                   "jmicron_raid_member", "lsi_mega_raid_member",
                   "nvidia_raid_member", "promise_fasttrack_raid_member",
                   "silicon_medley_raid_member", "via_raid_member"]
    _supported = False                  # is supported
    _hidden = True                      # hide devices with this formatting?

    def create(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        raise DMRaidMemberError("creation of dmraid members is non-sense")

    def destroy(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        raise DMRaidMemberError("destruction of dmraid members is non-sense")


register_device_format(DMRaidMember)



# events/manager.py
# Event management classes.
#
# Copyright (C) 2015-2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import abc
import inspect
from threading import current_thread, RLock, Thread
import pyudev
import six
import sys
import time
import traceback

from .. import threads
from .. import udev
from .. import util
from ..errors import EventManagerError, EventParamError
from ..flags import flags

from .changes import disable_callbacks, enable_callbacks

import logging
event_log = logging.getLogger("blivet.event")


def validate_cb(cb, kwargs=None, arg_count=None):
    """ Validate signature of callback function, returning True on success. """
    if not callable(cb):
        return False

    kwargs = kwargs or []
    arg_count = arg_count or 0

    if six.PY2:
        argspec = inspect.getargspec(cb)  # pylint: disable=deprecated-method,no-member
        if argspec.varargs or argspec.keywords:
            return True
        params = argspec.args
    else:
        params = list(inspect.signature(cb).parameters.keys())

    if len(params) < arg_count:
        return False

    return all(kw in params for kw in kwargs)


#
# Event
#
class Event(util.ObjectID):
    """ An external event. """
    def __init__(self, action, device, info=None):
        """
            :param str action: a string describing the type of event
            :param str device: (friendly) basename of device event operated on
            :param info: information about the device
        """
        self.initialized = time.time()
        self.action = action
        self.device = device
        self.info = info

    def __str__(self):
        return "%s %s [%d]" % (self.action, self.device, self.id)


class EventMask(util.ObjectID):
    """ Specification of events to ignore. """
    def __init__(self, device=None, action=None, partitions=False):
        """
            :keyword str device: basename of device to mask events on
            :keyword str action: action type to mask events of
            :keyword bool partitions: also match events on child partitions
        """
        self.device = device
        self.action = action
        self._partitions = partitions

    def _device_match(self, event):
        if self.device is None:
            return True

        if self.device == event.device:
            return True

        if (not self._partitions or
                not (udev.device_is_partition(event.info) or udev.device_is_dm_partition(event.info))):
            return False

        disk = udev.device_get_partition_disk(event.info)
        return disk and self.device == disk

    def _action_match(self, event):
        return self.action is None or self.action == event.action

    def match(self, event):
        """ Return True if this mask applies to the specified event.

            ..note::

                A mask whose device is a partitioned disk will match events
                on its partitions.
        """
        return self._device_match(event) and self._action_match(event)


#
# EventManager
#
@six.add_metaclass(abc.ABCMeta)
class EventManager(object):
    def __init__(self, handler_cb=None, notify_cb=None, error_cb=None):
        self._handler_cb = None
        """ event handler (must accept 'event', 'notify_cb' kwargs """

        self._notify_cb = None
        """ notification callback (must accept 'event', 'changes' kwargs) """

        self._error_cb = None
        """ exception handler (must be callable with one argument) """

        if handler_cb is not None:
            self.handler_cb = handler_cb

        if notify_cb is not None:
            self.notify_cb = notify_cb

        if error_cb is not None:
            self.error_cb = error_cb

        self._mask_list = list()
        """List of masks specifying events that should be ignored."""

        self._lock = RLock()
        """Re-entrant lock to serialize access to mask list."""

    @property
    def handler_cb(self):
        """ the main event handler """
        return self._handler_cb

    @handler_cb.setter
    def handler_cb(self, cb):
        if not validate_cb(cb, kwargs=["event", "notify_cb"]):
            raise EventParamError("invalid handler callback specified")

        self._handler_cb = cb

    @property
    def notify_cb(self):
        """ notification handler that runs after the main event handler """
        return self._notify_cb

    @notify_cb.setter
    def notify_cb(self, cb):
        if not validate_cb(cb, kwargs=["event", "changes"]):
            raise EventParamError("invalid notification callback specified")

        self._notify_cb = cb

    @property
    def error_cb(self):
        """ callback to run when an exception occurrs in a thread. """
        return self._error_cb

    @error_cb.setter
    def error_cb(self, cb):
        if not validate_cb(cb, arg_count=1):
            raise EventParamError("invalid error callback specified")

        self._error_cb = cb

    @property
    @abc.abstractmethod
    def enabled(self):
        return False

    @abc.abstractmethod
    def enable(self):
        """ Enable monitoring and handling of events.

            :raises: :class:`~.errors.EventManagerError` if no callback defined
        """
        if self.handler_cb is None:
            raise EventManagerError("cannot enable handler with no callback")

        event_log.info("enabling event handling")
        enable_callbacks()

    @abc.abstractmethod
    def disable(self):
        """ Disable monitoring and handling of events. """
        event_log.info("disabling event handling")
        disable_callbacks()

    def _mask_event(self, event):
        """ Return True if this event should be ignored """
        with self._lock:
            return six.next((m for m in self._mask_list if m.match(event)), None) is not None

    def add_mask(self, device=None, action=None, partitions=False):
        """ Add an event mask and return the new :class:`EventMask`.

            :keyword str device: ignore events on the named device
            :keyword str action: ignore events of the specified type
            :keyword bool partitions: also match events on child partitions

            device of None means mask events on all devices
            action of None means mask all event types
        """
        em = EventMask(device=device, action=action, partitions=partitions)
        with self._lock:
            self._mask_list.append(em)
        return em

    def remove_mask(self, mask):
        try:
            with self._lock:
                self._mask_list.remove(mask)
        except ValueError:
            pass

    @abc.abstractmethod
    def _create_event(self, *args, **kwargs):
        pass

    def handle_event(self, *args, **kwargs):
        """ Handle an event by running the registered handler.

            Currently the handler is run in a separate thread. This removes any
            threading-related expectations about the behavior of whatever is
            telling us about the events.

            Unhandled exceptions in event handler threads present a bit of a
            challenge. Generally, an unhandled exception in an event handler
            thread should be fatal because of the high likelihood that the
            error left the :class:`~.DeviceTree` in an inconsistent state.
            Since exceptions that occur in threads are not reported to the
            main thread, we have to try to notify it ourselves. Ideally, the
            calling application will have registered an :attr:`error_cb`,
            which we can use to notify it of the exception. If no callback
            has been set we can only wait until the next time the calling
            application makes a call into blivet, at which point we can
            raise the original exception.
        """
        event = self._create_event(*args, **kwargs)
        event_log.debug("new event: %s", event)

        if self._mask_event(event):
            event_log.debug("ignoring masked event %s", event)
            return

        t = Thread(target=self._run_event_handler,
                   name="event%d" % event.id,
                   kwargs={"event": event})
        t.daemon = True  # py2 compat
        t.start()

    def _run_event_handler(self, event):
        """ Run the event handler and account for unhandled exceptions. """
        if self.handler_cb is None:
            return

        try:
            # Pass the notify callback to the handler so it can run the
            # callback and pass thread-local data to it.
            self.handler_cb(event=event, notify_cb=self.notify_cb)  # pylint: disable=not-callable
        except Exception:  # pylint: disable=broad-except
            event_log.error(traceback.format_exc())
            exc_info = sys.exc_info()
            if self.error_cb is not None:
                self.error_cb(exc_info)  # pylint: disable=not-callable
            else:
                threads.save_thread_exception(current_thread(), exc_info)


class UdevEventManager(EventManager):
    def __init__(self, handler_cb=None, notify_cb=None):
        super(UdevEventManager, self).__init__(handler_cb=handler_cb, notify_cb=notify_cb)
        self._pyudev_observer = None

    @property
    def enabled(self):
        return self._pyudev_observer and self._pyudev_observer.monitor.started

    def enable(self):
        """ Enable monitoring and handling of block device uevents. """
        super(UdevEventManager, self).enable()
        monitor = pyudev.Monitor.from_netlink(udev.global_udev)
        monitor.filter_by("block")
        self._pyudev_observer = pyudev.MonitorObserver(monitor,
                                                       callback=self.handle_event,
                                                       name="monitor")
        self._pyudev_observer.start()
        with threads.blivet_lock:
            flags.uevents = True

    def disable(self):
        """ Disable monitoring and handling of block device uevents. """
        super(UdevEventManager, self).disable()
        if self.enabled:
            self._pyudev_observer.stop()

        self._pyudev_observer = None
        with threads.blivet_lock:
            flags.uevents = False

    def __call__(self, *args, **kwargs):
        return self

    def _create_event(self, *args, **kwargs):
        return Event(args[0].action, udev.device_get_name(args[0]), args[0])


event_manager = UdevEventManager()



# flags.py
#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import shlex


class Flags(object):

    def __init__(self):
        #
        # mode of operation
        #
        self.testing = False
        self.debug = False

        #
        # minor modes
        #
        self.uevents = False

        #
        # enable/disable functionality
        #
        try:
            import selinux
        except ImportError:
            self.selinux = False
        else:
            self.selinux = selinux.is_selinux_enabled()

        self.ibft = True
        self.noiswmd = False

        self.gfs2 = True
        self.jfs = True
        self.reiserfs = True

        # for this flag to take effect,
        # blockdev.mpath.set_friendly_names(flags.multipath_friendly_names) must
        # be called prior to calling Blivet.reset() or DeviceTree.populate()
        self.multipath_friendly_names = True

        # set to False since automatic updates of a device's information
        # or state should not be necessary by default
        self.auto_dev_updates = False

        # set to False by default since a forced reset for file contexts
        # is ordinary not necessary
        self.selinux_reset_fcon = False

        # set to True since we want to keep these around by default
        self.keep_empty_ext_partitions = True

        # set to False to suppress the default LVM behavior of saving
        # backup metadata in /etc/lvm/{archive,backup}
        self.lvm_metadata_backup = True

        # whether to include nodev filesystems in the devicetree
        self.include_nodev = False

        # whether to enable discard for newly created devices
        # (so far only for LUKS)
        self.discard_new = False

        self.boot_cmdline = {}

        self.update_from_boot_cmdline()
        self.allow_imperfect_devices = True

        # compression option for btrfs filesystems
        self.btrfs_compression = None

        self.debug_threads = False

        # Assign GPT partition type UUIDs to allow partition
        # auto-discovery according to:
        # https://uapi-group.org/specifications/specs/discoverable_partitions_specification/
        self.gpt_discoverable_partitions = False

    def get_boot_cmdline(self):
        with open("/proc/cmdline") as f:
            buf = f.read().strip()
            args = shlex.split(buf)
            for arg in args:
                (opt, _equals, val) = arg.partition("=")
                if val:
                    self.boot_cmdline[opt] = val

    def update_from_boot_cmdline(self):
        self.get_boot_cmdline()
        self.multipath = "nompath" not in self.boot_cmdline
        self.noiswmd = "noiswmd" in self.boot_cmdline
        self.gfs2 = "gfs2" in self.boot_cmdline
        self.jfs = "jfs" in self.boot_cmdline
        self.reiserfs = "reiserfs" in self.boot_cmdline


flags = Flags()



# populator/helpers/multipath.py
# Multipath backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from ... import udev
from ...devices import MultipathDevice
from ...storage_log import log_method_call
from .devicepopulator import DevicePopulator
from .formatpopulator import FormatPopulator

import logging
log = logging.getLogger("blivet")


class MultipathDevicePopulator(DevicePopulator):
    @classmethod
    def match(cls, data):
        return (udev.device_is_dm_mpath(data) and
                not udev.device_is_dm_partition(data))

    def run(self):
        name = udev.device_get_name(self.data)
        log_method_call(self, name=name)

        parent_devices = self._devicetree._add_parent_devices(self.data)

        device = None
        if parent_devices:
            device = MultipathDevice(name, parents=parent_devices,
                                     sysfs_path=udev.device_get_sysfs_path(self.data),
                                     wwn=parent_devices[0].wwn)
            self._devicetree._add_device(device)

        return device


class MultipathFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "multipath_member"

    def _get_kwargs(self):
        kwargs = super(MultipathFormatPopulator, self)._get_kwargs()
        # blkid does not care that the UUID it sees on a multipath member is
        # for the multipath set's (and not the member's) formatting, so we
        # have to discard it.
        kwargs.pop("uuid")
        kwargs.pop("label")
        return kwargs



# populator/helpers/disk.py
# Disk backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")
gi.require_version("GLib", "2.0")

from gi.repository import BlockDev as blockdev
from gi.repository import GLib

from ... import udev
from ... import util
from ...devices import DASDDevice, DiskDevice, FcoeDiskDevice, iScsiDiskDevice
from ...devices import MDBiosRaidArrayDevice, ZFCPDiskDevice, NVDIMMNamespaceDevice
from ...devices import NVMeNamespaceDevice, NVMeFabricsNamespaceDevice
from ...devices import device_path_to_name
from ...storage_log import log_method_call
from .devicepopulator import DevicePopulator

import logging
log = logging.getLogger("blivet")


class DiskDevicePopulator(DevicePopulator):
    priority = 10
    _device_class = DiskDevice

    @classmethod
    def match(cls, data):
        return udev.device_is_disk(data)

    def _get_kwargs(self):
        sysfs_path = udev.device_get_sysfs_path(self.data)
        kwargs = {
            "sysfs_path": sysfs_path,
            "serial": udev.device_get_serial(self.data),
            "vendor": util.get_sysfs_attr(sysfs_path, "device/vendor"),
            "model": util.get_sysfs_attr(sysfs_path, "device/model"),
            "bus": udev.device_get_bus(self.data),
            "wwn": udev.device_get_wwn(self.data)
        }

        if self._device_class == DiskDevice:
            kwargs["major"] = udev.device_get_major(self.data)
            kwargs["minor"] = udev.device_get_minor(self.data)
            log.info("%s is a disk", udev.device_get_name(self.data))

        return kwargs

    def run(self):
        name = udev.device_get_name(self.data)
        log_method_call(self, name=name)

        kwargs = self._get_kwargs()
        if not kwargs:
            return
        device = self._device_class(name, **kwargs)
        self._devicetree._add_device(device)
        return device


class iScsiDevicePopulator(DiskDevicePopulator):
    priority = 20
    _device_class = iScsiDiskDevice

    @classmethod
    def match(cls, data):
        from ...iscsi import iscsi
        return (super(iScsiDevicePopulator, iScsiDevicePopulator).match(data) and
                udev.device_is_iscsi(data) and iscsi.initiator and
                iscsi.initiator == udev.device_get_iscsi_initiator(data))

    def _get_kwargs(self):
        from ...iscsi import iscsi
        kwargs = super(iScsiDevicePopulator, self)._get_kwargs()
        kwargs["initiator"] = udev.device_get_iscsi_initiator(self.data)
        kwargs["address"] = udev.device_get_iscsi_address(self.data)
        kwargs["port"] = udev.device_get_iscsi_port(self.data)
        kwargs["target"] = udev.device_get_iscsi_name(self.data)
        kwargs["lun"] = udev.device_get_iscsi_lun(self.data)
        kwargs["name"] = udev.device_get_name(self.data)
        kwargs["iface"] = udev.device_get_iscsi_nic(self.data)
        kwargs["offload"] = kwargs["initiator"] != iscsi.initiator
        kwargs["id_path"] = udev.device_get_path(self.data)
        log.info("%s is an iscsi disk", kwargs["name"])

        # Backward compatibility attributes - to be removed
        if kwargs["offload"]:
            # qla4xxx partial offload
            kwargs["node"] = None
            kwargs["ibft"] = False
            kwargs["nic"] = "offload:not_accessible_via_iscsiadm"
        else:
            node = iscsi.get_node(kwargs["target"],
                                  kwargs["address"],
                                  kwargs["port"],
                                  kwargs["iface"])
            kwargs["node"] = node
            kwargs["ibft"] = node in iscsi.ibft_nodes
            if node:
                kwargs["nic"] = iscsi.ifaces.get(node.iface, node.iface)
            else:
                kwargs["nic"] = ""

        return kwargs


class FCoEDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = FcoeDiskDevice

    @classmethod
    def match(cls, data):
        return (super(FCoEDevicePopulator, FCoEDevicePopulator).match(data) and
                udev.device_is_fcoe(data))

    def _get_kwargs(self):
        kwargs = super(FCoEDevicePopulator, self)._get_kwargs()
        kwargs["nic"] = udev.device_get_fcoe_nic(self.data)
        kwargs["identifier"] = udev.device_get_fcoe_identifier(self.data)
        kwargs["id_path"] = udev.device_get_path(self.data)
        log.info("%s is an fcoe disk", udev.device_get_name(self.data))
        return kwargs


class MDBiosRaidDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = MDBiosRaidArrayDevice

    @classmethod
    def match(cls, data):
        return (super(MDBiosRaidDevicePopulator, MDBiosRaidDevicePopulator).match(data) and
                udev.device_get_md_container(data))

    def _get_kwargs(self):
        kwargs = super(MDBiosRaidDevicePopulator, self)._get_kwargs()
        parent_path = udev.device_get_md_container(self.data)
        parent_name = device_path_to_name(parent_path)
        container = self._devicetree.get_device_by_name(parent_name)

        # FIXME: Move this whole block to an add_parent_devices method or similar
        if not container:
            parent_sys_name = blockdev.md.node_from_name(parent_name)
            container_sysfs = "/sys/class/block/" + parent_sys_name
            container_info = udev.get_device(container_sysfs)
            if not container_info:
                log.error("failed to find md container %s at %s",
                          parent_name, container_sysfs)
                return

            self._devicetree.handle_device(container_info)
            container = self._devicetree.get_device_by_name(parent_name)
            if not container:
                log.error("failed to scan md container %s", parent_name)
                return

        kwargs["parents"] = [container]
        kwargs["level"] = udev.device_get_md_level(self.data)
        kwargs["member_devices"] = udev.device_get_md_devices(self.data)
        kwargs["uuid"] = udev.device_get_md_uuid(self.data)
        kwargs["exists"] = True
        # remove some kwargs that don't make sense for md
        del kwargs["model"]
        del kwargs["serial"]
        del kwargs["vendor"]
        del kwargs["bus"]
        del kwargs["wwn"]
        return kwargs


class DASDDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = DASDDevice

    @classmethod
    def match(cls, data):
        return (super(DASDDevicePopulator, DASDDevicePopulator).match(data) and
                udev.device_is_dasd(data))

    def _get_kwargs(self):
        kwargs = super(DASDDevicePopulator, self)._get_kwargs()
        kwargs["busid"] = udev.device_get_dasd_bus_id(self.data)
        kwargs["opts"] = {}
        for attr in ['readonly', 'use_diag', 'erplog', 'failfast']:
            kwargs["opts"][attr] = udev.device_get_dasd_flag(self.data, attr)

        log.info("%s is a dasd device", udev.device_get_name(self.data))
        return kwargs


class ZFCPDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = ZFCPDiskDevice

    @classmethod
    def match(cls, data):
        return (super(ZFCPDevicePopulator, ZFCPDevicePopulator).match(data) and
                udev.device_is_zfcp(data))

    def _get_kwargs(self):
        kwargs = super(ZFCPDevicePopulator, self)._get_kwargs()

        for attr in ['hba_id', 'wwpn', 'fcp_lun']:
            kwargs[attr] = udev.device_get_zfcp_attribute(self.data, attr=attr)

        log.info("%s is a zfcp device", udev.device_get_name(self.data))
        return kwargs


class NVDIMMNamespaceDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = NVDIMMNamespaceDevice

    @classmethod
    def match(cls, data):
        return (super(NVDIMMNamespaceDevicePopulator, NVDIMMNamespaceDevicePopulator).match(data) and
                udev.device_is_nvdimm_namespace(data))

    def _get_kwargs(self):
        kwargs = super(NVDIMMNamespaceDevicePopulator, self)._get_kwargs()

        from ...static_data import nvdimm
        ninfo = nvdimm.get_namespace_info(self.data.get("DEVNAME"))

        kwargs["mode"] = blockdev.nvdimm_namespace_get_mode_str(ninfo.mode)
        kwargs["devname"] = ninfo.dev
        kwargs["uuid"] = ninfo.uuid
        kwargs["sector_size"] = ninfo.sector_size
        kwargs["id_path"] = udev.device_get_path(self.data)

        log.info("%s is an NVDIMM namespace device", udev.device_get_name(self.data))
        return kwargs


class NVMeNamespaceDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = NVMeNamespaceDevice

    @classmethod
    def match(cls, data):
        return (super(NVMeNamespaceDevicePopulator, NVMeNamespaceDevicePopulator).match(data) and
                udev.device_is_nvme_namespace(data) and not udev.device_is_nvme_fabrics(data))

    def _get_kwargs(self):
        kwargs = super(NVMeNamespaceDevicePopulator, self)._get_kwargs()

        log.info("%s is an NVMe local namespace device", udev.device_get_name(self.data))

        if not hasattr(blockdev.Plugin, "NVME"):
            # the nvme plugin is not generally available
            return kwargs

        path = udev.device_get_devname(self.data)
        try:
            ninfo = blockdev.nvme_get_namespace_info(path)
        except GLib.GError as err:
            log.debug("Failed to get namespace info for %s: %s", path, str(err))
        else:
            kwargs["nsid"] = ninfo.nsid
            kwargs["uuid"] = ninfo.uuid
            kwargs["eui64"] = ninfo.eui64
            kwargs["nguid"] = ninfo.nguid

        log.info("%s is an NVMe local namespace device", udev.device_get_name(self.data))
        return kwargs


class NVMeFabricsNamespaceDevicePopulator(DiskDevicePopulator):
    priority = 20

    _device_class = NVMeFabricsNamespaceDevice

    @classmethod
    def match(cls, data):
        return (super(NVMeFabricsNamespaceDevicePopulator, NVMeFabricsNamespaceDevicePopulator).match(data) and
                udev.device_is_nvme_namespace(data) and udev.device_is_nvme_fabrics(data))

    def _get_kwargs(self):
        kwargs = super(NVMeFabricsNamespaceDevicePopulator, self)._get_kwargs()

        log.info("%s is an NVMe fabrics namespace device", udev.device_get_name(self.data))

        if not hasattr(blockdev.Plugin, "NVME"):
            # the nvme plugin is not generally available
            return kwargs

        path = udev.device_get_devname(self.data)
        try:
            ninfo = blockdev.nvme_get_namespace_info(path)
        except GLib.GError as err:
            log.debug("Failed to get namespace info for %s: %s", path, str(err))
        else:
            kwargs["nsid"] = ninfo.nsid

        return kwargs



# devicetree.py
# Device management for anaconda's storage configuration module.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import os
import pprint
import re
import six

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from .actionlist import ActionList
from .callbacks import callbacks
from .errors import DeviceError, DeviceTreeError, StorageError, DuplicateUUIDError, InvalidMultideviceSelection
from .deviceaction import ActionDestroyDevice, ActionDestroyFormat
from .devices import BTRFSDevice, NoDevice, PartitionDevice
from .devices import LVMLogicalVolumeDevice, LVMVolumeGroupDevice
from .devices.lib import Tags
from . import formats
from .devicelibs import lvm
from .events.handler import EventHandlerMixin
from . import util
from .populator import PopulatorMixin
from .storage_log import log_method_call, log_method_return
from .threads import SynchronizedMeta
from .static_data import lvs_info

import logging
log = logging.getLogger("blivet")

_LVM_DEVICE_CLASSES = (LVMLogicalVolumeDevice, LVMVolumeGroupDevice)


@six.add_metaclass(SynchronizedMeta)
class DeviceTreeBase(object):
    """ A quasi-tree that represents the devices in the system.

        The tree contains a list of :class:`~.devices.StorageDevice` instances,
        which does not necessarily reflect the actual state of the system's
        devices. :class:`~.deviceaction.DeviceAction` is used to perform
        modifications to the tree, except when initially populating the tree.

        :class:`~.deviceaction.DeviceAction` instances are registered, possibly
        causing the addition or removal of :class:`~.devices.StorageDevice`
        instances to/from the tree. A :class:`~.deviceaction.DeviceAction`
        is reversible up to the time its 'execute' method is called.

        Only one action of any given type/object pair should exist for
        any given device at any given time.

        :class:`~.deviceaction.DeviceAction` instances can only be registered
        for leaf devices, except for resize actions.
    """
    def __init__(self, ignored_disks=None, exclusive_disks=None):
        """
            :keyword ignored_disks: ignored disks
            :type ignored_disks: list
            :keyword exclusive_disks: exclusive didks
            :type exclusive_disks: list
        """
        self.reset(ignored_disks, exclusive_disks)

    def reset(self, ignored_disks=None, exclusive_disks=None):
        """ Reset the instance to its initial state.

            :keyword ignored_disks: ignored disks
            :type ignored_disks: list
            :keyword exclusive_disks: exclusive didks
            :type exclusive_disks: list
        """
        # internal data members
        self._devices = []
        self._actions = ActionList(addfunc=self._register_action,
                                   removefunc=self._cancel_action)

        self._hidden = []

        lvm.lvm_devices_reset()

        self.exclusive_disks = exclusive_disks or []
        self.ignored_disks = ignored_disks or []

        self.edd_dict = {}

    def __str__(self):
        done = []

        def show_subtree(root, depth):
            abbreviate_subtree = root in done
            s = "%s%s\n" % ("  " * depth, root)
            done.append(root)
            if abbreviate_subtree:
                s += "%s...\n" % ("  " * (depth + 1),)
            else:
                for child in root.children:
                    s += show_subtree(child, depth + 1)
            return s

        roots = [d for d in self._devices if not d.parents]
        tree = ""
        for root in roots:
            tree += show_subtree(root, 0)
        return tree

    #
    # Device list
    #
    @property
    def devices(self):
        """ List of devices currently in the tree """
        devices = []
        for device in self._devices:
            if not getattr(device, "complete", True):
                continue

            if device.uuid and device.uuid in [d.uuid for d in devices] and \
               not isinstance(device, NoDevice):
                raise DeviceTreeError("duplicate uuids in device tree")

            devices.append(device)

        return devices

    @property
    def names(self):
        """ List of devices names """
        lv_info = list(lvs_info.cache.keys())

        names = []
        for dev in self._devices + self._hidden:
            # don't include "req%d" partition names
            if (dev.type != "partition" or not dev.name.startswith("req")) and \
               dev.type != "btrfs volume" and \
               dev.name not in names:
                names.append(dev.name)

        # include LVs that are not in the devicetree and not scheduled for removal
        removed_names = [ac.device.name for ac in self.actions.find(action_type="destroy",
                                                                    object_type="device")]
        names.extend(n for n in lv_info if n not in names and n not in removed_names)

        return names

    def _add_device(self, newdev, new=True):
        """ Add a device to the tree.

            :param newdev: the device to add
            :type newdev: a subclass of :class:`~.devices.StorageDevice`

            Raise DeviceTreeError if the device's identifier is already
            in the list.
        """
        if newdev.uuid and newdev.uuid in [d.uuid for d in self._devices] and \
           not isinstance(newdev, NoDevice):
            # Just found a device with already existing UUID. Is it the same device?
            dev = self.get_device_by_uuid(newdev.uuid, incomplete=True, hidden=True)
            if dev.name == newdev.name:
                raise DeviceTreeError("Trying to add already existing device.")
            else:
                raise DuplicateUUIDError("Duplicate UUID '%s' found for devices: "
                                         "'%s' and '%s'." % (newdev.uuid, newdev.name, dev.name))

        # make sure this device's parent devices are in the tree already
        for parent in newdev.parents:
            if parent not in self._devices:
                raise DeviceTreeError("parent device not in tree")

        newdev.add_hook(new=new)
        self._devices.append(newdev)

        callbacks.device_added(device=newdev)
        log.info("added %s %s (id %d) to device tree", newdev.type,
                 newdev.name,
                 newdev.id)

    def _remove_device(self, dev, force=None, modparent=True):
        """ Remove a device from the tree.

            :param dev: the device to remove
            :type dev: a subclass of :class:`~.devices.StorageDevice`
            :keyword force: whether to force removal of a non-leaf device
            :type force: bool
            :keyword modparent: update parent device to account for removal
            :type modparent: bool

            .. note::

                Only leaves may be removed.
        """
        if dev not in self._devices:
            raise ValueError("Device '%s' not in tree" % dev.name)

        if not dev.isleaf and not force:
            log.debug("%s has children %s", dev.name, pprint.pformat(c.name for c in dev.children))
            raise ValueError("Cannot remove non-leaf device '%s'" % dev.name)

        dev.remove_hook(modparent=modparent)
        if modparent:
            # if this is a partition we need to remove it from the parted.Disk
            if isinstance(dev, PartitionDevice) and dev.disk is not None:
                # adjust all other PartitionDevice instances belonging to the
                # same disk so the device name matches the potentially altered
                # name of the parted.Partition
                for device in self._devices:
                    if isinstance(device, PartitionDevice) and \
                       device.disk == dev.disk:
                        device.update_name()

        self._devices.remove(dev)
        callbacks.device_removed(device=dev)
        log.info("removed %s %s (id %d) from device tree", dev.type,
                 dev.name,
                 dev.id)

    def recursive_remove(self, device, actions=True, remove_device=True, modparent=True):
        """ Remove a device after removing its dependent devices.

            :param :class:`~.devices.StorageDevice` device: the device to remove
            :keyword bool actions: whether to schedule actions for the removal
            :keyword bool modparent: whether to update parent device upon removal
            :keyword bool remove_device: whether to remove the root device

            If the device is not a leaf, all of its dependents are removed
            recursively until it is a leaf device. At that point the device is
            removed, unless it is a disk. If the device is a disk, its
            formatting is removed but no attempt is made to actually remove the
            disk device.
        """
        log.debug("removing %s", device.name)
        devices = self.get_dependent_devices(device)

        # this isn't strictly necessary, but it makes the action list easier to
        # read when removing logical partitions because of the automatic
        # renumbering that happens if you remove them in ascending numerical
        # order
        devices.reverse()

        while devices:
            log.debug("devices to remove: %s", [d.name for d in devices])
            leaves = [d for d in devices if d.isleaf]
            log.debug("leaves to remove: %s", [d.name for d in leaves])
            for leaf in leaves:
                if actions:
                    if leaf.format.exists and not leaf.protected and \
                       not leaf.format_immutable:
                        self.actions.add(ActionDestroyFormat(leaf))

                    self.actions.add(ActionDestroyDevice(leaf))
                else:
                    if not leaf.format_immutable:
                        leaf.format = None
                    self._remove_device(leaf, modparent=modparent)

                devices.remove(leaf)

        if not device.format_immutable:
            if actions:
                self.actions.add(ActionDestroyFormat(device))
            else:
                device.format = None

        if remove_device and not device.is_disk:
            if actions:
                self.actions.add(ActionDestroyDevice(device))
            else:
                self._remove_device(device, modparent=modparent)

    #
    # Actions
    #
    @property
    def actions(self):
        return self._actions

    def _register_action(self, action):
        """ Register an action to be performed at a later time.

            :param action: the action
            :type action: :class:`~.deviceaction.DeviceAction`

            Modifications to the Device instance are handled before we
            get here.
        """
        if not (action.is_create and action.is_device) and \
           action.device not in self._devices:
            raise DeviceTreeError("device is not in the tree")
        elif (action.is_create and action.is_device):
            if action.device in self._devices:
                raise DeviceTreeError("device is already in the tree")

        if action.is_create and action.is_device:
            # if adding an LV constructed from other LVs, we need to remove the
            # LVs it's supposed to be constructed from the device tree
            if isinstance(action.device, LVMLogicalVolumeDevice) and action.device.from_lvs:
                for lv in action.device.from_lvs:
                    if lv in self._devices:
                        self._remove_device(lv)
            self._add_device(action.device)
        elif action.is_destroy and action.is_device:
            self._remove_device(action.device)
            # if removing an LV constructed from other LVs, we need to put the
            # LVs it's supposed to be constructed from back into the device tree
            if isinstance(action.device, LVMLogicalVolumeDevice) and action.device.from_lvs:
                for lv in action.device.from_lvs:
                    self._add_device(lv, new=False)
        elif action.is_create and action.is_format:
            if isinstance(action.device.format, formats.fs.FS) and \
               action.device.format.mountpoint in self.filesystems:
                raise DeviceTreeError("mountpoint already in use")

    def _cancel_action(self, action):
        """ Cancel a registered action.

            :param action: the action
            :type action: :class:`~.deviceaction.DeviceAction`

            This will unregister the action and do any required
            modifications to the device list.

            Actions all operate on a Device, so we can use the devices
            to determine dependencies.
        """
        if action.is_create and action.is_device:
            # remove the device from the tree
            self._remove_device(action.device)
            if isinstance(action.device, LVMLogicalVolumeDevice) and action.device.from_lvs:
                # if removing an LV constructed from other LVs, we need to put the
                # LVs it's supposed to be constructed from back into the device tree
                for lv in action.device.from_lvs:
                    self._add_device(lv, new=False)
        elif action.is_destroy and action.is_device:
            # if adding an LV constructed from other LVs, we need to remove the
            # LVs it's supposed to be constructed from the device tree
            if isinstance(action.device, LVMLogicalVolumeDevice) and action.device.from_lvs:
                for lv in action.device.from_lvs:
                    if lv in self._devices:
                        self._remove_device(lv)
            # add the device back into the tree
            self._add_device(action.device, new=False)

    #
    # Device control
    #
    def teardown_all(self):
        """ Run teardown methods on all devices. """
        for device in self.leaves:
            if device.protected:
                continue

            try:
                device.teardown(recursive=True)
            except (StorageError, blockdev.BlockDevError) as e:
                log.info("teardown of %s failed: %s", device.name, e)

    def setup_all(self):
        """ Run setup methods on all devices. """
        for device in self.leaves:
            try:
                device.setup()
            except DeviceError as e:
                log.error("setup of %s failed: %s", device.name, e)

    #
    # Device search by relation
    #
    def get_dependent_devices(self, dep, hidden=False):
        """ Return a list of devices that depend on dep.

            The list includes both direct and indirect dependents.

            :param dep: the device whose dependents we are looking for
            :type dep: :class:`~.devices.StorageDevice`
            :keyword bool hidden: include hidden devices in search
        """
        dependents = []
        log_method_call(self, dep=dep, hidden=hidden)

        # don't bother looping looking for dependents if this is a leaf device
        # XXX all hidden devices are leaves
        if dep.isleaf and not hidden:
            log.debug("dep is a leaf")
            return dependents

        devices = self._devices[:]
        if hidden:
            devices.extend(self._hidden)

        for device in devices:
            log.debug("checking if %s depends on %s", device.name, dep.name)
            if device.depends_on(dep):
                dependents.append(device)

        return dependents

    def get_related_disks(self, disk):
        """ Return disks related to disk by container membership.

            :param :class:`~.devices.StorageDevice` disk: the disk
            :returns: related disks
            :rtype: set of :class:`~.devices.StorageDevice`

            .. note::

                The disk may be hidden.

        """
        return set(d for dep in self.get_dependent_devices(disk, hidden=True)
                   for d in dep.disks)

    def get_disk_actions(self, disks):
        """ Return a list of actions related to the specified disk.

            :param disks: list of disks
            :type disk: list of :class:`~.devices.StorageDevices`
            :returns: list of related actions
            :rtype: list of :class:`~.deviceaction.DeviceAction`

            This includes all actions on the specified disks, plus all actions
            on disks that are in any way connected to the specified disk via
            container devices.
        """
        # This is different from get_related_disks in that we are finding disks
        # related by any action -- not just the current state of the devicetree.
        related_disks = set()
        for action in self.actions:
            if any(action.device.depends_on(d) for d in disks):
                related_disks.update(action.device.disks)

        # now related_disks could be a superset of disks, so go through and
        # build a list of actions related to any disk in related_disks
        # Note that this list preserves the ordering of the action list.
        related_actions = [a for a in self.actions
                           if set(a.device.disks).intersection(related_disks)]
        return related_actions

    def cancel_disk_actions(self, disks):
        """ Cancel all actions related to the specified disk.

            :param disks: list of disks
            :type disk: list of :class:`~.devices.StorageDevices`

            This includes actions related directly and indirectly (via container
            membership, for example).
        """
        actions = self.get_disk_actions(disks)
        for action in reversed(actions):
            self.actions.remove(action)

    #
    # Device search by property
    #
    def _filter_devices(self, incomplete=False, hidden=False):
        """ Return list of devices modified according to parameters.

            :param bool incomplete: include incomplete devices in result
            :param bool hidden: include hidden devices in result

            :returns: a generator of devices
            :rtype: generator of :class:`~.devices.Device`
        """
        if hidden:
            devices = (d for d in self._devices[:] + self._hidden[:])
        else:
            devices = (d for d in self._devices[:])

        if not incomplete:
            devices = (d for d in devices if getattr(d, "complete", True))
        return devices

    def get_device_by_sysfs_path(self, path, incomplete=False, hidden=False):
        """ Return a list of devices with a matching sysfs path.

            :param str path: the sysfs path to match
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, path=path, incomplete=incomplete, hidden=hidden)
        result = None
        if path:
            devices = self._filter_devices(incomplete=incomplete, hidden=hidden)
            result = six.next((d for d in devices if d.sysfs_path == path), None)
        log_method_return(self, result)
        return result

    def get_device_by_uuid(self, uuid, incomplete=False, hidden=False):
        """ Return a list of devices with a matching UUID.

            :param str uuid: the UUID to match
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, uuid=uuid, incomplete=incomplete, hidden=hidden)
        result = None
        if uuid:
            devices = self._filter_devices(incomplete=incomplete, hidden=hidden)
            result = six.next((d for d in devices if d.uuid == uuid or d.format.uuid == uuid), None)
        log_method_return(self, result)
        return result

    def get_device_by_label(self, label, incomplete=False, hidden=False):
        """ Return a device with a matching filesystem label.

            :param str label: the filesystem label to match
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, label=label, incomplete=incomplete, hidden=hidden)
        result = None
        if label:
            devices = self._filter_devices(incomplete=incomplete, hidden=hidden)
            result = six.next((d for d in devices if getattr(d.format, "label", None) == label), None)
        log_method_return(self, result)
        return result

    def get_device_by_name(self, name, incomplete=False, hidden=False):
        """ Return a device with a matching name.

            :param str name: the name to look for
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, name=name, incomplete=incomplete, hidden=hidden)
        result = None
        if name:
            devices = self._filter_devices(incomplete=incomplete, hidden=hidden)
            result = six.next((d for d in devices if d.name == name or
                               (isinstance(d, _LVM_DEVICE_CLASSES) and d.name == name.replace("--", "-"))),
                              None)
        log_method_return(self, result)
        return result

    def get_device_by_path(self, path, incomplete=False, hidden=False):
        """ Return a device with a matching path.

            If there is more than one device with a matching path,
            prefer a leaf device to a non-leaf device.

            :param str path: the path to match
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, path=path, incomplete=incomplete, hidden=hidden)
        result = None
        if path:
            devices = self._filter_devices(incomplete=incomplete, hidden=hidden)

            # The usual order of the devices list is one where leaves are at
            # the end. So that the search can prefer leaves to interior nodes
            # the list that is searched is the reverse of the devices list.
            result = six.next((d for d in reversed(list(devices)) if d.path == path or
                               (isinstance(d, _LVM_DEVICE_CLASSES) and d.path == path.replace("--", "-"))),
                              None)

        log_method_return(self, result)
        return result

    def get_device_by_id(self, id_num, incomplete=False, hidden=False):
        """ Return a device with specified device id.

            :param int id_num: the id to look for
            :param bool incomplete: include incomplete devices in search
            :param bool hidden: include hidden devices in search
            :returns: the first matching device found
            :rtype: :class:`~.devices.Device`
        """
        log_method_call(self, id_num=id_num, incomplete=incomplete, hidden=hidden)
        devices = self._filter_devices(incomplete=incomplete, hidden=hidden)
        result = six.next((d for d in devices if d.id == id_num), None)
        log_method_return(self, result)
        return result

    def resolve_device(self, devspec, blkid_tab=None, crypt_tab=None, options=None):
        """ Return the device matching the provided device specification.

            The spec can be anything from a device name (eg: 'sda3') to a device
            node path (eg: '/dev/mapper/fedora-root' or '/dev/dm-2') to
            something like 'UUID=xyz-tuv-qrs' or 'LABEL=rootfs'.

            :param devspec: a string describing a block device
            :type devspec: str
            :keyword blkid_tab: blkid info
            :type blkid_tab: :class:`~.BlkidTab`
            :keyword crypt_tab: crypto info
            :type crypt_tab: :class:`~.CryptTab`
            :keyword options: mount options
            :type options: str
            :returns: the device
            :rtype: :class:`~.devices.StorageDevice` or None
        """
        # find device in the tree
        device = None
        if devspec.startswith("UUID="):
            # device-by-uuid
            uuid = devspec.partition("=")[2]
            if ((uuid.startswith('"') and uuid.endswith('"')) or
                    (uuid.startswith("'") and uuid.endswith("'"))):
                uuid = uuid[1:-1]
            device = self.uuids.get(uuid)
        elif devspec.startswith("LABEL="):
            # device-by-label
            label = devspec.partition("=")[2]
            if ((label.startswith('"') and label.endswith('"')) or
                    (label.startswith("'") and label.endswith("'"))):
                label = label[1:-1]
            device = self.labels.get(label)
        elif options and "nodev" in options.split(","):
            device = self.get_device_by_name(devspec)
            if not device:
                device = self.get_device_by_path(devspec)
        else:
            if re.match(r'(0x)?[A-Fa-f0-9]{2}(p\d+)?$', devspec):
                # BIOS drive number
                (drive, _p, partnum) = devspec.partition("p")
                spec = int(drive, 16)
                for (edd_name, edd_number) in self.edd_dict.items():
                    if edd_number == spec:
                        device = self.get_device_by_name(edd_name + partnum)
                        break
            if not device and not devspec.startswith("/dev/"):
                device = self.get_device_by_name(devspec)
                if not device:
                    devspec = "/dev/" + devspec

            if not device:
                if devspec.startswith("/dev/disk/"):
                    devspec = os.path.realpath(devspec)

                if devspec.startswith("/dev/dm-"):
                    try:
                        dm_name = blockdev.dm.name_from_node(devspec[5:])
                    except blockdev.DMError as e:
                        log.info("failed to resolve %s: %s", devspec, e)
                        dm_name = None

                    if dm_name:
                        devspec = "/dev/mapper/" + dm_name

                if re.match(r'/dev/md\d+(p\d+)?$', devspec):
                    try:
                        md_name = blockdev.md.name_from_node(devspec[5:])
                    except blockdev.MDRaidError as e:
                        log.info("failed to resolve %s: %s", devspec, e)
                        md_name = None

                    if md_name:
                        devspec = "/dev/md/" + md_name

                # device path
                device = self.get_device_by_path(devspec)

            if device is None:
                if blkid_tab:
                    # try to use the blkid.tab to correlate the device
                    # path with a UUID
                    blkid_tab_ent = blkid_tab.get(devspec)
                    if blkid_tab_ent:
                        log.debug("found blkid.tab entry for '%s'", devspec)
                        uuid = blkid_tab_ent.get("UUID")
                        if uuid:
                            device = self.get_device_by_uuid(uuid)
                            if device:
                                devstr = device.name
                            else:
                                devstr = "None"
                            log.debug("found device '%s' in tree", devstr)
                        if device and device.format and \
                           device.format.type == "luks":
                            map_name = device.format.map_name
                            log.debug("luks device; map name is '%s'", map_name)
                            mapped_dev = self.get_device_by_name(map_name)
                            if mapped_dev:
                                device = mapped_dev

                if device is None and crypt_tab and \
                   devspec.startswith("/dev/mapper/"):
                    # try to use a dm-crypt mapping name to
                    # obtain the underlying device, possibly
                    # using blkid.tab
                    crypt_tab_ent = crypt_tab.get(devspec.split("/")[-1])
                    if crypt_tab_ent:
                        luks_dev = crypt_tab_ent['device']
                        try:
                            device = luks_dev.children[0]
                        except IndexError as e:
                            pass
                elif device is None:
                    # dear lvm: can we please have a few more device nodes
                    #           for each logical volume?
                    #           three just doesn't seem like enough.
                    name = devspec[5:]      # strip off leading "/dev/"

                    (vg_name, _slash, lv_name) = name.partition("/")
                    if lv_name and "/" not in lv_name:
                        # looks like we may have one
                        lv = "%s-%s" % (vg_name, lv_name)
                        device = self.get_device_by_name(lv)

        # check mount options for btrfs volumes in case it's a subvol
        if device and device.type.startswith("btrfs") and options:
            # start with the volume -- not a subvolume
            device = getattr(device, "volume", device)

            attr = None
            if "subvol=" in options:
                attr = "name"
                val = util.get_option_value("subvol", options)
            elif "subvolid=" in options:
                attr = "vol_id"
                val = util.get_option_value("subvolid", options)
            elif device.default_subvolume:
                # default subvolume
                device = device.default_subvolume

            if attr and val:
                for subvol in device.subvolumes:
                    if getattr(subvol, attr, None) == val:
                        device = subvol
                        break

        if device:
            log.debug("resolved '%s' to '%s' (%s)", devspec, device.name, device.type)
        else:
            log.debug("failed to resolve '%s'", devspec)
        return device

    #
    # Conveniences
    #
    @property
    def leaves(self):
        """ List of all devices upon which no other devices exist. """
        leaves = [d for d in self._devices if d.isleaf]
        return leaves

    @property
    def filesystems(self):
        """ List of filesystems. """
        filesystems = []
        for dev in self.leaves:
            if dev.format and getattr(dev.format, 'mountpoint', None):
                filesystems.append(dev.format)

        return filesystems

    @property
    def uuids(self):
        """ Dict with uuid keys and :class:`~.devices.Device` values. """
        uuids = {}
        for dev in self._devices:
            try:
                uuid = dev.uuid
            except AttributeError:
                uuid = None

            if uuid:
                uuids[uuid] = dev

            try:
                uuid = dev.format.uuid
            except AttributeError:
                uuid = None

            if uuid:
                uuids[uuid] = dev

        return uuids

    @property
    def labels(self):
        """ Dict with label keys and Device values.

            FIXME: duplicate labels are a possibility
        """
        labels = {}
        for dev in self._devices:
            # don't include btrfs member devices
            if getattr(dev.format, "label", None) and \
               (dev.format.type != "btrfs" or isinstance(dev, BTRFSDevice)):
                labels[dev.format.label] = dev

        return labels

    @property
    def mountpoints(self):
        """ Dict with mountpoint keys and Device values. """
        filesystems = {}
        for device in self.devices:
            if device.format.mountable and device.format.mountpoint:
                filesystems[device.format.mountpoint] = device
        return filesystems

    #
    # Disk filter
    #
    def hide(self, device):
        """ Hide the specified device.

            :param device: the device to hide
            :type device: :class:`~.devices.StorageDevice`

            Hiding a device will cancel all actions that involve the device and
            will remove the device from the device list.

            If the device is not a leaf device, all devices that depend on it
            will be hidden leaves-first until the device is a leaf device.

            If a device exists, performs some special actions and places
            it on a list of hidden devices.

            Mixes recursion and side effects, most significantly in the code
            that removes all the actions. However, this code is a null op
            in every case except the first base case that is reached,
            where all actions are removed. This means that when a device
            is removed explicitly in this function by means of a direct call to
            _remove_devices it is guaranteed that all actions have already
            been canceled.

            If a device does not exist then it must have been removed by the
            cancelation of all the actions, so it does not need to be removed
            explicitly.

            Most devices are considered leaf devices if they have no children,
            however, some devices must satisfy more stringent requirements.
            _remove_device() will raise an exception if the device it is
            removing is not a leaf device. hide() guarantees that any
            device that it removes will have no children, but it does not
            guarantee that the more stringent requirements will be enforced.
            Therefore, _remove_device() is invoked with the force parameter
            set to True, to skip the isleaf check.
        """
        if device in self._hidden:
            return

        # cancel actions first thing so that we hide the correct set of devices
        if device.is_disk:
            # Cancel all actions on this disk and any disk related by way of an
            # aggregate/container device (eg: lvm volume group).
            self.cancel_disk_actions([device])

        for d in device.children:
            self.hide(d)

        log.info("hiding device %s", device)

        if not device.exists:
            return

        self._remove_device(device, force=True, modparent=False)

        self._hidden.append(device)
        if device.format.type == "lvmpv":
            lvm.lvm_devices_remove(device.path)

    def unhide(self, device):
        """ Restore a device's visibility.

            :param device: the device to restore/unhide
            :type device: :class:`~.devices.StorageDevice`

            .. note::

                Actions canceled while hiding the device are not rescheduled
                automatically.

        """

        # the hidden list should be in leaves-first order
        for hidden in reversed(self._hidden):
            if hidden == device or hidden.depends_on(device) and \
               not any(parent in self._hidden for parent in hidden.parents):

                log.info("unhiding device %s %s (id %d)", hidden.type,
                         hidden.name,
                         hidden.id)
                self._hidden.remove(hidden)
                self._devices.append(hidden)
                hidden.add_hook(new=False)
                if hidden.format.type == "lvmpv":
                    lvm.lvm_devices_add(hidden.path)

    def expand_taglist(self, taglist):
        """ Expands tags in input list into devices.

            :param taglist: list of strings
            :returns: set of devices
            :rtype: set of strings

            Raise ValueError if unknown tag encountered in taglist

            .. note::

                Returns empty set if taglist is empty
        """
        result = set()
        devices = self._devices[:]
        for item in taglist:
            if item.startswith('@'):
                tag = item[1:]
                if tag not in Tags.__members__:
                    raise ValueError("unknown tag '@%s' encountered" % tag)
                for device in devices:
                    if tag in device.tags:
                        result.add(device.name)
            else:
                result.add(item)
        return result

    def _disk_in_taglist(self, disk, taglist):
        # Taglist is a list containing mix of disk names and tags into which disk may belong.
        # Check if it does. Raise ValueError if unknown tag is encountered.
        if disk.name in taglist:
            return True
        tags = [t[1:] for t in taglist if t.startswith("@")]
        for tag in tags:
            if tag not in Tags.__members__:
                raise ValueError("unknown ignoredisk tag '@%s' encountered" % tag)
            if Tags(tag) in disk.tags:
                return True
        return False

    def _is_ignored_disk(self, disk):
        """ Checks config for lists of exclusive and ignored disks
            and returns if the given one should be ignored
        """
        return ((self.ignored_disks and self._disk_in_taglist(disk, self.ignored_disks)) or
                (self.exclusive_disks and not self._disk_in_taglist(disk, self.exclusive_disks)))

    def _hide_ignored_disks(self):
        # hide any subtrees that begin with an ignored disk
        for disk in [d for d in self._devices if d.is_disk]:
            is_ignored = self.ignored_disks and self._disk_in_taglist(disk, self.ignored_disks)
            is_exclusive = self.exclusive_disks and self._disk_in_taglist(disk, self.exclusive_disks)

            if is_ignored:
                if len(disk.children) == 1:
                    if not all(self._is_ignored_disk(d) for d in disk.children[0].parents):
                        raise InvalidMultideviceSelection("Including only a subset of raid/multipath member disks is not allowed.")

                    # and also children like fwraid or mpath
                    self.hide(disk.children[0])

                # this disk is ignored: ignore it and all it's potential parents
                for p in disk.parents:
                    self.hide(p)

                # and finally hide the disk itself
                self.hide(disk)

            if self.exclusive_disks and not is_exclusive:
                ignored = True
                # If the filter allows all members of a fwraid or mpath, the
                # fwraid or mpath itself is implicitly allowed as well. I don't
                # like this very much but we have supported this usage in the
                # past, so I guess we will support it forever.
                if disk.parents and all(p.format.hidden for p in disk.parents):
                    ignored = any(self._is_ignored_disk(d) for d in disk.parents)
                elif disk.format.hidden and len(disk.children) == 1:
                    # Similarly, if the filter allows an mpath or fwraid, we cannot
                    # ignore the member devices.
                    ignored = self._is_ignored_disk(disk.children[0])

                if ignored:
                    self.hide(disk)


class DeviceTree(DeviceTreeBase, PopulatorMixin, EventHandlerMixin):
    def __init__(self, ignored_disks=None, exclusive_disks=None, disk_images=None):
        DeviceTreeBase.__init__(self, ignored_disks=ignored_disks, exclusive_disks=exclusive_disks)
        PopulatorMixin.__init__(self, disk_images=disk_images)
        EventHandlerMixin.__init__(self)

    # pylint: disable=arguments-differ
    def reset(self, ignored_disks=None, exclusive_disks=None, disk_images=None):
        DeviceTreeBase.reset(self, ignored_disks=ignored_disks, exclusive_disks=exclusive_disks)
        PopulatorMixin.reset(self, disk_images=disk_images)



# devices/__init__.py
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from .lib import device_path_to_name, device_name_to_disk_by_path, ParentList
from .device import Device
from .storage import StorageDevice
from .disk import DiskDevice, DiskFile, MultipathDevice, iScsiDiskDevice, FcoeDiskDevice, DASDDevice, ZFCPDiskDevice, NVDIMMNamespaceDevice, NVMeNamespaceDevice, NVMeFabricsNamespaceDevice
from .partition import PartitionDevice
from .dm import DMDevice, DMLinearDevice, DMCryptDevice, DMIntegrityDevice, DM_MAJORS
from .luks import LUKSDevice, IntegrityDevice
from .lvm import LVMVolumeGroupDevice, LVMLogicalVolumeDevice
from .md import MDBiosRaidArrayDevice, MDContainerDevice, MDRaidArrayDevice, MD_MAJORS
from .btrfs import BTRFSDevice, BTRFSVolumeDevice, BTRFSSubVolumeDevice, BTRFSSnapShotDevice
from .file import FileDevice, DirectoryDevice, SparseFileDevice
from .loop import LoopDevice
from .optical import OpticalDevice
from .nodev import NoDevice, TmpFSDevice
from .network import NetworkStorageDevice
from .nfs import NFSDevice
from .stratis import StratisPoolDevice, StratisFilesystemDevice



#
# edd.py
# BIOS EDD data parsing functions
#
# Copyright 2010-2015 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s):
#            Peter Jones <pjones@redhat.com>
#            Hans de Goede <hdegoede@redhat.com>
#            Ales Kozumplik <akozumpl@redhat.com>
#

import logging
import os
import re
import struct
import copy

from .. import util

log = logging.getLogger("blivet")
testdata_log = logging.getLogger("testdata")
testdata_log.setLevel(logging.DEBUG)

re_bios_device_number = re.compile(r'.*/int13_dev([0-9a-fA-F]+)/*$')
re_host_bus_pci = re.compile(r'^(PCIX|PCI|XPRS|HTPT)\s*(\S*)\s*channel: (\S*)\s*$')
re_interface_atapi = re.compile(r'^ATAPI\s*device: (\S*)\s*lun: (\S*)\s*$')
re_interface_ata = re.compile(r'^ATA\s*device: (\S*)\s*$')
re_interface_scsi = re.compile(r'^SCSI\s*id: (\S*)\s*lun: (\S*)\s*$')
re_interface_usb = re.compile(r'^USB\s*serial_number: (\S*)\s*$')
re_interface_1394 = re.compile(r'^1394\s*eui: (\S*)\s*$')
re_interface_fibre = re.compile(r'^FIBRE\s*wwid: (\S*)\s*lun: (\S*)\s*$')
re_interface_i2o = re.compile(r'^I2O\s*identity_tag: (\S*)\s*$')
# pretty sure the RAID definition using "identity_tag" is basically a kernel
# bug, but it's part of the ABI now, so it sticks.  The format of the
# scnprintf() is at least correct.
re_interface_raid = re.compile(r'^RAID\s*identity_tag: (\S*)\s*$')
re_interface_edd3_sata = re.compile(r'^SATA\s*device: (\S*)\s*$')
# EDD 4 features from 2010 and later.  Awesomely, the "version" output from
# int 13 AH=41h says: AH Version of extensions. Shall be set to 30h,
# so there's no way to distinguish these from EDD 3, even thuogh SATA does
# differ.  In theory, if we're on <4.0, pmp should always be all 0's.
re_interface_edd4_sata = re.compile(r'^SATA\s*device: (\S*)\s*pmp: (\S*)\s*$')
re_interface_sas = re.compile(r'^SAS\s*sas_address: (\S*)\s*lun: \(\S*\)\s*$')
# to make life difficult, when it finds an unknown interface type string,
# the kernel prints the values without the string.  But it does print the
# anchoring tab that would go between them...
re_interface_unknown = re.compile(r'^(\S*)\s*unknown: (\S*) (\S*)\s*$')


class EddEntry(object):

    """ This object merely collects what the /sys/firmware/edd/* entries can
        provide.
    """

    def __init__(self, sysfspath, root=None):
        self.root = util.Path(root or "", root="")

        # some misc data from various files...
        self.sysfspath = util.Path(sysfspath, root=self.root)
        """ sysfspath is the path we're probing
        """

        match = re_bios_device_number.match(sysfspath)
        self.bios_device_number = int(match.group(1), base=16)
        """ The device number from the EDD path """

        self.sysfslink = None
        """ The path /sys/block/BLAH is a symlink link to once we've resolved
            that this is a particular device.  Used for logging later.
        """

        self.version = util.get_sysfs_attr(self.sysfspath, "version")
        """ The edd version this entry claims conformance with, from
            /sys/firmware/edd/int13_devXX/version """

        self.mbr_sig = None
        """ The MBR signature data from edd/int13_devXX/mbr_signature """

        self.sectors = None
        """ The number of sectors on the device from edd/int13_devXX/sectors """

        # Now the data from edd/int13_devXX/host_bus
        self.host_bus = None
        """ The ID string for the host bus type, from
            edd/int13_devXX/host_bus.
        """
        self.pci_dev = None
        """ The host bus bus:device.function, from edd/int13_devXX/host_bus.
        """
        self.channel = None
        """ The host bus device's channel number edd/int13_devXX/host_bus.
            The spec says:
              Channel number. If more than one interface of the same type is
              accessed through a single Bus, Slot, Function, then the channel
              number shall identify each interface. If there is only one
              interface, the content of this field shall be cleared to zero. If
              there are two interfaces, such as an ATA Primary and Secondary
              interface, the primary interface shall be zero, and the secondary
              interface shall be one.

              Values 00h through FEh shall represent a valid Channel Number.

              Value FFh shall indicate that this field is not used.

              If the device is connected to a SATA controller functioning in
              non-PATA emulation mode, this byte shall be FFh.
        """

        # And now the various data from different formats of
        # edd/int13_devXX/interface .
        self.interface = None
        """ interface is the actual contents of the interface file,
            preserved for logging later.
        """

        self.type = None
        """ The device type from edd/int13_devXX/interface.
        """

        self.atapi_device = None
        """ The device number of the ATAPI device from
            edd/int13_devXX/interface when self.type is ATAPI.
        """
        self.atapi_lun = None
        """ The LUN of the ATAPI device from edd/int13_devXX/interface when
            self.type is ATAPI.
        """

        self.ata_device = None
        """ The device number from edd/int13_devXX/interface when self.type
            is ATA or SATA (because Linux treats these the same.)
        """
        self.ata_pmp = None
        """ The ATA port multiplier ID from edd/int13_devXX/interface when
            self.type is SATA.
        """

        self.scsi_id = None
        """ The SCSI device ID from edd/int13_devXX/interface when
            self.type is SCSI
        """
        self.scsi_lun = None
        """ The SCSI device LUN from edd/int13_devXX/interface when
            self.type is SCSI
        """

        self.usb_serial = None
        """ The USB storage device's serial number from
            edd/int13_devXX/interface when self.type is USB.
        """

        self.ieee1394_eui64 = None
        """ The Firewire/IEEE-1394 EUI-64 ID from edd/int13_devXX/interface
            when self.type is 1394.
        """

        self.fibre_wwid = None
        """ The FibreChannel WWID from edd/int13_devXX/interface when
            self.type is FIBRE.
        """
        self.fibre_lun = None
        """ The FibreChannel LUN from edd/int13_devXX/interface when
            self.type is FIBRE.
        """

        self.i2o_identity = None
        """ The I2O Identity from edd/int13_devXX/interface when self.type
            is I2O.
        """

        self.sas_address = None
        """ The SAS Address from edd/int13_devXX/interface when self.type
            is SAS.
        """
        self.sas_lun = None
        """ The SAS LUN from edd/int13_devXX/interface when self.type is SAS.
        """

        self.load()

    def _fmt(self, line_pad, separator):
        s = "%(t)spath: %(sysfspath)s version: %(version)s %(nl)s" \
            "%(t)smbr_signature: %(mbr_sig)s sectors: %(sectors)s"
        if self.type is not None:
            s += " %(type)s"
        if self.sysfslink is not None:
            s += "%(nl)s%(t)ssysfs pci path: %(sysfslink)s"
        if any([self.host_bus, self.pci_dev, self.channel is not None]):
            s += "%(nl)s%(t)shost_bus: %(host_bus)s pci_dev: %(pci_dev)s "\
                "channel: %(channel)s"
        if self.interface is not None:
            s += "%(nl)s%(t)sinterface: \"%(interface)s\""
        if any([self.atapi_device is not None, self.atapi_lun is not None]):
            s += "%(nl)s%(t)satapi_device: %(atapi_device)s " \
                 "atapi_lun: %(atapi_lun)s"
        if self.ata_device is not None:
            s += "%(nl)s%(t)sata_device: %(ata_device)s"
            if self.ata_pmp is not None:
                s += ", ata_pmp: %(ata_pmp)s"
        if any([self.scsi_id is not None, self.scsi_lun is not None]):
            s += "%(nl)s%(t)sscsi_id: %(scsi_id)s, scsi_lun: %(scsi_lun)s"
        if self.usb_serial is not None:
            s += "%(nl)s%(t)susb_serial: %(usb_serial)s"
        if self.ieee1394_eui64 is not None:
            s += "%(nl)s%(t)s1394_eui: %(ieee1394_eui64)s"
        if any([self.fibre_wwid, self.fibre_lun]):
            s += "%(nl)s%(t)sfibre wwid: %(fibre_wwid)s lun: %(fibre_lun)s"
        if self.i2o_identity is not None:
            s += "%(nl)s%(t)si2o_identity: %(i2o_identity)s"
        if any([self.sas_address, self.sas_lun]):
            s += "%(nl)s%(t)ssas_address: %(sas_address)s sas_lun: %(sas_lun)s"

        d = copy.copy(self.__dict__)
        d['t'] = line_pad
        d['nl'] = separator

        return s % d

    def __gt__(self, other):
        if not isinstance(self, other.__class__) and \
           not isinstance(other, self.__class__):
            return self.__class__ > other.__class__
        ldict = copy.copy(self.__dict__)
        rdict = copy.copy(other.__dict__)
        del ldict["root"]
        del rdict["root"]
        return ldict > rdict

    def __eq__(self, other):
        if not isinstance(self, other.__class__) and \
           not isinstance(other, self.__class__):
            return self.__class__ == other.__class__
        ldict = copy.copy(self.__dict__)
        rdict = copy.copy(other.__dict__)
        del ldict["root"]
        del rdict["root"]
        return ldict == rdict

    def __lt__(self, other):
        if not isinstance(self, other.__class__) and \
           not isinstance(other, self.__class__):
            return self.__class__ < other.__class__
        ldict = copy.copy(self.__dict__)
        rdict = copy.copy(other.__dict__)
        del ldict["root"]
        del rdict["root"]
        return ldict < rdict

    def __str__(self):
        return self._fmt('\t', '\n')

    def __repr__(self):
        return "<EddEntry%s>" % (self._fmt(' ', ''),)

    def __getitem__(self, idx):
        return str(self)[idx]

    def __len__(self):
        return len(str(self))

    def load(self):
        interface = util.get_sysfs_attr(self.sysfspath, "interface")
        # save this so we can log it from the matcher.
        self.interface = interface
        if interface:
            try:
                self.type = interface.split()[0]
                if self.type == "ATAPI":
                    match = re_interface_atapi.match(interface)
                    self.atapi_device = int(match.group(1))
                    self.atapi_lun = int(match.group(2))
                elif self.type == "ATA":
                    match = re_interface_ata.match(interface)
                    self.ata_device = int(match.group(1))
                elif self.type == "SCSI":
                    match = re_interface_scsi.match(interface)
                    self.scsi_id = int(match.group(1))
                    self.scsi_lun = int(match.group(2))
                elif self.type == "USB":
                    match = re_interface_usb.match(interface)
                    self.usb_serial = int(match.group(1), base=16)
                elif self.type == "1394":
                    match = re_interface_1394.match(interface)
                    self.ieee1394_eui64 = int(match.group(1), base=16)
                elif self.type == "FIBRE":
                    match = re_interface_fibre.match(interface)
                    self.fibre_wwid = int(match.group(1), base=16)
                    self.fibre_lun = int(match.group(2), base=16)
                elif self.type == "I2O":
                    match = re_interface_i2o.match(interface)
                    self.i2o_identity = int(match.group(1), base=16)
                elif self.type == "RAID":
                    match = re_interface_raid.match(interface)
                    self.raid_array = int(match.group(1), base=16)
                elif self.type == "SATA":
                    match = re_interface_edd4_sata.match(interface)
                    if match:
                        self.ata_device = int(match.group(1))
                        self.ata_pmp = int(match.group(2))
                    else:
                        match = re_interface_edd3_sata.match(interface)
                        self.ata_device = int(match.group(1))
                elif self.type == "SAS":
                    sas_match = re_interface_sas.match(interface)
                    unknown_match = re_interface_unknown.match(interface)
                    if sas_match:
                        self.sas_address = int(sas_match.group(1), base=16)
                        self.sas_lun = int(sas_match.group(2), base=16)
                    elif unknown_match:
                        self.sas_address = int(unknown_match.group(2), base=16)
                        self.sas_lun = int(unknown_match.group(3), base=16)
                    else:
                        log.warning("edd: can not match interface for %s: %s",
                                    self.sysfspath, interface)
                else:
                    log.warning("edd: can not match interface for %s: %s",
                                self.sysfspath, interface)
            except AttributeError as e:
                if e.args == "'NoneType' object has no attribute 'group'":
                    log.warning("edd: can not match interface for %s: %s",
                                self.sysfspath, interface)
                else:
                    raise e

        self.mbr_sig = util.get_sysfs_attr(self.sysfspath, "mbr_signature")
        sectors = util.get_sysfs_attr(self.sysfspath, "sectors")
        if sectors:
            self.sectors = int(sectors)
        hbus = util.get_sysfs_attr(self.sysfspath, "host_bus")
        if hbus:
            match = re_host_bus_pci.match(hbus)
            if match:
                self.host_bus = match.group(1)
                self.pci_dev = match.group(2)
                self.channel = int(match.group(3))
            else:
                log.warning("edd: can not match host_bus for %s: %s",
                            self.sysfspath, hbus)


class EddMatcher(object):

    """ This object tries to match given entry to a disk device name.

        Assuming, heuristic analysis and guessing hapens here.
    """

    def __init__(self, edd_entry, root=None):
        self.edd = edd_entry
        self.root = root or ""

    def devname_from_ata_pci_dev(self):
        pattern = util.Path('/sys/block/*', root=self.root)
        retries = []

        def match_port(components, ata_port, ata_port_idx, path, link):
            fn = util.Path(util.join_paths(components[0:6] +
                                           ['ata_port', ata_port]), root=self.root)
            port_no = int(util.get_sysfs_attr(fn, 'port_no'))

            if self.edd.type == "ATA":
                # On ATA, port_no is kernel's ata_port->local_port_no, which
                # should be the same as the ata device number.
                if port_no != self.edd.ata_device:
                    return
            else:
                # On SATA, "port_no" is the kernel's ata_port->print_id, which
                # is awesomely ata_port->id + 1, where ata_port->id is edd's
                # ata_device
                if port_no != self.edd.ata_device + 1:
                    return

            fn = components[0:6] + ['link%d' % (ata_port_idx,), ]
            exp = [r'.*'] + fn + [r'dev%d\.(\d+)(\.(\d+)){0,1}$' % (ata_port_idx,)]
            exp = util.join_paths(exp)
            expmatcher = re.compile(exp)

            pmp = util.join_paths(fn + ['dev%d.*.*' % (ata_port_idx,)])
            pmp = util.Path(pmp, root=self.root)
            dev = util.join_paths(fn + ['dev%d.*' % (ata_port_idx,)])
            dev = util.Path(dev, root=self.root)
            for ataglob in [pmp, dev]:
                for atapath in ataglob.glob():
                    match = expmatcher.match(atapath.ondisk)
                    if match is None:
                        continue

                    # so at this point it's devW.X.Y or devW.Z as such:
                    # dev_set_name(dev, "dev%d.%d",
                    # ap->print_id,ata_dev->devno); dev_set_name(dev,
                    # "dev%d.%d.0", ap->print_id, link->pmp); we care about
                    # print_id and pmp for matching and the ATA channel if
                    # applicable. We already checked print_id above.
                    if match.group(3) is None:
                        channel = int(match.group(1))
                        if (self.edd.channel == 255 and channel == 0) or \
                                (self.edd.channel == channel):
                            yield ({'link': util.Path(link, root=self.root),
                                    'path': path.split('/')[-1]})
                    else:
                        pmp = int(match.group(1))
                        if self.edd.ata_pmp == pmp:
                            yield ({'link': util.Path(link, root=self.root),
                                    'path': path.split('/')[-1]})

        answers = []
        for path in pattern.glob():
            emptyslash = util.Path("/", root=self.root)
            path = util.Path(path, root=self.root)
            link = util.sysfs_readlink(path=emptyslash, link=path)
            testdata_log.debug("sysfs link: \"%s\" -> \"%s\"", path, link)
            # just add /sys/block/ at the beginning so it's always valid
            # paths in the filesystem...
            components = ['/sys/block'] + link.split('/')
            if len(components) != 11:
                continue
            # ATA and SATA paths look like:
            # ../devices/pci0000:00/0000:00:1f.2/ata1/host0/target0:0:0/0:0:0:0/block/sda
            # where literally the only pieces of data here are
            # "pci0000:00:1f.2", "ata1", and "sda".
            #
            # EDD 3's "channel" doesn't really mean anything at all on SATA,
            # and 255 means "not in use".  Any other value should be an ATA
            # device (but might be a SATA device in compat mode), and that
            # matches N in devM.N .  So basically "channel" means master/slave
            # for ATA (non-SATA) devices.  Also in EDD 3, SATA port multipliers
            # aren't represented in any way.
            #
            # In EDD 4, which unfortunately says to leave 0x30 as the version
            # number, the port multiplier id is an additional field on the
            # interface.  So basically we should try to use the value the
            # kernel gives us*, but we can't trust it.  Thankfully there
            # won't be a devX.Y.Z (i.e. a port multiplier device) in sysfs
            # that collides with devX.Z (a non-port-multiplied device),
            # so if we get a value from the kernel, we can try with and
            # without it.
            #
            # * When the kernel finally learns of these facts...
            #
            if components[4] != '0000:%s' % (self.edd.pci_dev,):
                continue
            if not components[5].startswith('ata'):
                continue
            ata_port = components[5]
            ata_port_idx = int(components[5][3:])

            # strictly this should always be required, but #!@#!@#!@ seabios
            # iterates the sata device number /independently/ of the host
            # bridge it claims things are attached to.  In that case this
            # the scsi target will always have "0" as the ID component.
            args = {'device': self.edd.ata_device}
            exp = r"target\d+:0:%(device)s/\d+:0:%(device)s:0/block/.*" % args
            matcher = re.compile(exp)
            match = matcher.match("/".join(components[7:]))
            if not match:
                retries.append({
                    'components': components,
                    'ata_port': ata_port,
                    'ata_port_idx': ata_port_idx,
                    'path': path,
                    'link': link,
                })
                continue

            for port in match_port(components, ata_port, ata_port_idx, path,
                                   link):
                answers.append(port)

        # now handle the ones we discarded because libata's scsi id doesn't
        # match the ata_device.
        for retry in retries:
            for port in match_port(**retry):
                if answers:
                    log.warning("edd: ignoring possible extra match for ATA device %s channel %s ata %d pmp %s: %s",
                                self.edd.pci_dev, self.edd.channel,
                                self.edd.ata_device, self.edd.ata_pmp,
                                retry['path'])
                else:
                    log.warning("edd: using possible extra match for ATA device %s channel %s ata %d pmp %s: %s",
                                self.edd.pci_dev, self.edd.channel,
                                self.edd.ata_device, self.edd.ata_pmp,
                                retry['path'])
                    answers.append(port)

        if len(answers) > 1:
            log.error("edd: Found too many ATA devices for EDD device 0x%x: %s",
                      self.edd.bios_device_number,
                      [a['link'] for a in answers])
        if len(answers) > 0:
            self.edd.sysfslink = answers[0]['link']
            return answers[0]['path']
        else:
            log.warning(
                "edd: Could not find ATA device for pci dev %s channel %s ata %d pmp %s",
                self.edd.pci_dev, self.edd.channel,
                self.edd.ata_device, self.edd.ata_pmp)

        return None

    def devname_from_virtio_scsi_pci_dev(self):
        if self.edd.scsi_id is None or self.edd.scsi_lun is None:
            return None
        # Virtio SCSI looks like scsi but with a virtio%d/ stuck in the middle
        # channel appears to be a total lie on VirtIO SCSI devices.
        tmpl = "../devices/pci0000:00/0000:%(pci_dev)s/virtio*/" \
            "host*/target*:0:%(dev)d/*:0:%(dev)d:%(lun)d/block/"
        args = {
            'pci_dev': self.edd.pci_dev,
            'dev': self.edd.scsi_id,
            'lun': self.edd.scsi_lun,
        }
        pattern = util.Path(tmpl % args, root=self.root + "/sys/block/")
        answers = []
        for mp in pattern.glob():
            # Normal VirtIO devices just have the block link right there...
            block_entries = os.listdir(mp.ondisk)
            for be in block_entries:
                link = mp + be
                answers.append({'link': link, 'path': be})

        if len(answers) > 1:
            log.error("Found too many VirtIO SCSI devices for EDD device 0x%x: %s",
                      self.edd.bios_device_number,
                      [a['link'] for a in answers])
        if len(answers) > 0:
            self.edd.sysfslink = answers[0]['link']
            return answers[0]['path']
        else:
            log.info("edd: Could not find VirtIO SCSI device for pci dev %s "
                     "channel %s scsi id %s lun %s", self.edd.pci_dev,
                     self.edd.channel, self.edd.scsi_id, self.edd.scsi_lun)

    def devname_from_scsi_pci_dev(self):
        tmpl = "../devices/pci0000:00/0000:%(pci_dev)s/" \
            "host%(chan)d/target%(chan)d:0:%(dev)d/" \
            "%(chan)d:0:%(dev)d:%(lun)d/block/"
        args = {
            'pci_dev': self.edd.pci_dev,
            'chan': self.edd.channel,
            'dev': self.edd.scsi_id,
            'lun': self.edd.scsi_lun,
        }
        pattern = util.Path(tmpl % args, root=self.root + "/sys/block/")
        answers = []
        for mp in pattern.glob():
            # Normal VirtIO devices just have the block link right there...
            block_entries = os.listdir(mp.ondisk)
            for be in block_entries:
                link = mp + be
                answers.append({'link': link, 'path': be})

        if len(answers) > 1:
            log.error("Found too many SCSI devices for EDD device 0x%x: %s",
                      self.edd.bios_device_number,
                      [a['link'] for a in answers])
        if len(answers) > 0:
            self.edd.sysfslink = answers[0]['link']
            return answers[0]['path']
        else:
            log.warning("edd: Could not find SCSI device for pci dev %s "
                        "channel %s scsi id %s lun %s", self.edd.pci_dev,
                        self.edd.channel, self.edd.scsi_id, self.edd.scsi_lun)
        return None

    def devname_from_virt_pci_dev(self):
        pattern = util.Path("../devices/pci0000:00/0000:%s/virtio*/block/" %
                            (self.edd.pci_dev,), root=self.root + "/sys/block/")
        answers = []
        for mp in pattern.glob():
            # Normal VirtIO devices just have the block link right there...
            block_entries = os.listdir(mp.ondisk)
            for be in block_entries:
                link = mp + be
                answers.append({'link': link, 'path': be})

        if len(answers) > 1:
            log.error("Found too many VirtIO devices for EDD device 0x%x: %s",
                      self.edd.bios_device_number,
                      [a['link'] for a in answers])
        if len(answers) > 0:
            self.edd.sysfslink = answers[0]['link']
            return answers[0]['path']
        else:
            log.info(
                "edd: Could not find Virtio device for pci dev %s channel %s",
                self.edd.pci_dev, self.edd.channel)

        return None

    def devname_from_pci_dev(self):
        if self.edd.pci_dev is None:
            return None
        name = self.devname_from_virt_pci_dev()
        if name is not None:
            return name
        name = self.devname_from_virtio_scsi_pci_dev()
        if name is not None:
            return name

        unsupported = ("ATAPI", "USB", "1394", "I2O", "RAID", "FIBRE", "SAS")
        if self.edd.type in unsupported:
            log.warning("edd: interface type %s is not implemented (%s)",
                        self.edd.type, self.edd.sysfspath)
            log.warning("edd: interface details: %s", self.edd.interface)
        if self.edd.type in ("ATA", "SATA") and \
                self.edd.ata_device is not None:
            name = self.devname_from_ata_pci_dev()
        elif self.edd.type == "SCSI":
            name = self.devname_from_scsi_pci_dev()
        if self.edd.sysfslink:
            path = util.Path("/sys/block/", root=self.root) \
                + self.edd.sysfslink \
                + "/device"
            link = os.readlink(path.ondisk)
            testdata_log.debug("sysfs link: \"%s\" -> \"%s\"", path, link)

        return name

    def match_via_mbrsigs(self, mbr_dict):
        """ Try to match the edd entry based on its mbr signature.

            This will obviously fail for a fresh drive/image, but in extreme
            cases can also show false positives for randomly matching data.
        """
        sysblock = util.Path("/sys/block/", root=self.root)
        for (name, mbr_sig) in mbr_dict.items():
            if mbr_sig == self.edd.mbr_sig:
                self.edd.sysfslink = util.sysfs_readlink(sysblock, link=name)
                return name
        return None


def collect_edd_data(root=None):
    edd_data_dict = {}
    globstr = util.Path("/sys/firmware/edd/int13_dev*/", root=root)
    for path in globstr.glob():
        match = re_bios_device_number.match(path)
        biosdev = int("0x%s" % (match.group(1),), base=16)
        log.debug("edd: found device 0x%x at %s", biosdev, path)
        edd_data_dict[biosdev] = EddEntry(path, root=root)
    return edd_data_dict


def collect_mbrs(devices, root=None):
    """ Read MBR signatures from devices.

        Returns a dict mapping device names to their MBR signatures. It is not
        guaranteed this will succeed, with a new disk for instance.
    """
    mbr_dict = {}
    for dev in devices:
        fd = -1
        try:
            path = util.Path("/dev", root=root) + dev.name
            fd = os.open(path.ondisk, os.O_RDONLY)
            # The signature is the unsigned integer at byte 440:
            os.lseek(fd, 440, 0)
            data = os.read(fd, 4)
            mbrsig = struct.unpack('I', data)
            sdata = struct.unpack("BBBB", data)
            sdata = "".join(["%02x" % (x,) for x in sdata])
            os.close(fd)
            testdata_log.debug("device %s data[440:443] = %s", path, sdata)
        except OSError as e:
            testdata_log.debug("device %s data[440:443] raised %s", path, e)
            log.error("edd: could not read mbrsig from disk %s: %s",
                      dev.name, str(e))
            if fd > 0:
                os.close(fd)
            continue

        mbrsig_str = "0x%08x" % mbrsig
        # sanity check
        if mbrsig_str == '0x00000000':
            log.info("edd: MBR signature on %s is zero. new disk image?",
                     dev.name)
            continue
        else:
            for (dev_name, mbrsig_str_old) in mbr_dict.items():
                if mbrsig_str_old == mbrsig_str:
                    log.error("edd: dupicite MBR signature %s for %s and %s",
                              mbrsig_str, dev_name, dev.name)
                    # this actually makes all the other data useless
                    return {}
        # update the dictionary
        mbr_dict[dev.name] = mbrsig_str
    log.info("edd: collected mbr signatures: %s", mbr_dict)
    return mbr_dict


def get_edd_dict(devices, root=None):
    """ Generates the 'device name' -> 'edd number' mapping.

        The EDD kernel module that exposes /sys/firmware/edd is thoroughly
        broken, the information there is incomplete and sometimes downright
        wrong. So after we mine out all useful information that the files under
        /sys/firmware/edd/int13_*/ can provide, we resort to heuristics and
        guessing. Our first attempt is, by looking at the device type int
        'interface', attempting to map pci device number, channel number etc. to
        a sysfs path, check that the path really exists, then read the device
        name (e.g 'sda') from there. Should this fail we try to match contents
        of 'mbr_signature' to a real MBR signature found on the existing block
        devices.
    """
    mbr_dict = collect_mbrs(devices, root=root)
    edd_entries_dict = collect_edd_data(root=root)
    edd_dict = {}
    for (edd_number, edd_entry) in edd_entries_dict.items():
        matcher = EddMatcher(edd_entry, root=root)
        # first try to match through the pci dev etc.
        name = matcher.devname_from_pci_dev()
        log.debug("edd: data extracted from 0x%x:%r", edd_number, edd_entry)
        if name:
            log.info("edd: matched 0x%x to %s using PCI dev", edd_number, name)
        # next try to compare mbr signatures
        else:
            name = matcher.match_via_mbrsigs(mbr_dict)
            if name:
                log.info("edd: matched 0x%x to %s using MBR sig", edd_number, name)

        if name:
            old_edd_number = edd_dict.get(name)
            if old_edd_number:
                log.info("edd: both edd entries 0x%x and 0x%x seem to map to %s",
                         old_edd_number, edd_number, name)
                # this means all the other data can be confused and useless
                return {}
            edd_dict[name] = edd_number
        else:
            log.error("edd: unable to match edd entry 0x%x", edd_number)
    return edd_dict



# fslabeling.py
# Filesystem labeling classes for anaconda's storage configuration module.
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass

from . import availability


@add_metaclass(abc.ABCMeta)
class FSLabeling(object):

    """An abstract class that represents filesystem labeling actions.
    """

    default_label = abc.abstractproperty(
        doc="Default label set on this filesystem at creation.")

    @classmethod
    @abc.abstractmethod
    def label_format_ok(cls, label):
        """Returns True if this label is correctly formatted for this
           filesystem, otherwise False.

           :param str label: the label for this filesystem
           :rtype: bool
        """
        raise NotImplementedError


class Ext2FSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 17


class FATFSLabeling(FSLabeling):

    default_label = "" if availability.MKDOSFS_NEW_APP.available else "NO NAME"

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 12


class JFSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 17


class ReiserFSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 17


class XFSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return ' ' not in label and len(label) < 13


class HFSLabeling(FSLabeling):

    default_label = "Untitled"

    @classmethod
    def label_format_ok(cls, label):
        return ':' not in label and len(label) < 28 and len(label) > 0


class HFSPlusLabeling(FSLabeling):

    default_label = "Untitled"

    @classmethod
    def label_format_ok(cls, label):
        return ':' not in label and 0 < len(label) < 129


class NTFSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 129


class F2FSLabeling(FSLabeling):

    default_label = ""

    @classmethod
    def label_format_ok(cls, label):
        return len(label) < 513



# fswritelabel.py
# Filesystem label writing classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass

from .. import util
from ..errors import FSWriteLabelError

from . import availability
from . import fstask
from . import task


@add_metaclass(abc.ABCMeta)
class FSWriteLabel(task.BasicApplication, fstask.FSTask):

    """ An abstract class that represents writing a label for a filesystem. """

    description = "write filesystem label"

    args = abc.abstractproperty(doc="arguments for writing a label")

    # IMPLEMENTATION methods

    @property
    def _set_command(self):
        """Get the command to label the filesystem.

           :return: the command
           :rtype: list of str
        """
        return [str(self.ext)] + self.args

    def do_task(self):  # pylint: disable=arguments-differ
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSWriteLabelError("\n".join(error_msgs))

        rc = util.run_program(self._set_command)
        if rc:
            raise FSWriteLabelError("label failed")


class DosFSWriteLabel(FSWriteLabel):
    ext = availability.DOSFSLABEL_APP

    @property
    def args(self):
        if availability.MKDOSFS_NEW_APP.available:
            if self.fs.label:
                return [self.fs.device, self.fs.label]
            else:
                return [self.fs.device, "--reset"]
        else:
            return [self.fs.device, self.fs.label]


class Ext2FSWriteLabel(FSWriteLabel):
    ext = availability.E2LABEL_APP

    @property
    def args(self):
        return [self.fs.device, self.fs.label]


class JFSWriteLabel(FSWriteLabel):
    ext = availability.JFSTUNE_APP

    @property
    def args(self):
        return ["-L", self.fs.label, self.fs.device]


class NTFSWriteLabel(FSWriteLabel):
    ext = availability.NTFSLABEL_APP

    @property
    def args(self):
        return [self.fs.device, self.fs.label]


class ReiserFSWriteLabel(FSWriteLabel):
    ext = availability.REISERFSTUNE_APP

    @property
    def args(self):
        return ["-l", self.fs.label, self.fs.device]


class XFSWriteLabel(FSWriteLabel):
    ext = availability.XFSADMIN_APP

    @property
    def args(self):
        return ["-L", self.fs.label if self.fs.label != "" else "--", self.fs.device]


class UnimplementedFSWriteLabel(fstask.UnimplementedFSTask):
    pass



# swap.py
# Device format classes for anaconda's storage configuration module.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import os

from parted import PARTITION_SWAP, fileSystemType
from ..errors import FSWriteUUIDError, SwapSpaceError
from ..storage_log import log_method_call
from ..tasks import availability
from ..tasks import fsuuid
from . import DeviceFormat, register_device_format
from ..size import Size
from .. import udev

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")


class SwapSpace(DeviceFormat):

    """ Swap space """
    _type = "swap"
    _name = None
    _udev_types = ["swap"]
    parted_flag = PARTITION_SWAP
    parted_system = fileSystemType["linux-swap(v1)"]
    _formattable = True                # can be formatted
    _supported = True                  # is supported
    _linux_native = True                # for clearpart
    _plugin = availability.BLOCKDEV_SWAP_PLUGIN

    _max_size = Size("16 TiB")

    config_actions_map = {"label": "write_label"}

    def __init__(self, **kwargs):
        """
            :keyword device: path to the block device node
            :keyword uuid: this swap space's uuid
            :keyword exists: whether this is an existing format
            :type exists: bool
            :keyword label: this swap space's label
            :keyword priority: this swap space's priority
            :type priority: int

            .. note::

                The 'device' kwarg is required for existing formats. For non-
                existent formats, it is only necessary that the 'device'
                attribute be set before the :meth:`~.SwapSpace.create` method
                runs. You can specify the device at the last moment by via the
                'device' kwarg to the :meth:`~.SwapSpace.create` method.
        """
        log_method_call(self, **kwargs)
        DeviceFormat.__init__(self, **kwargs)

        self.priority = kwargs.get("priority", -1)
        self.label = kwargs.get("label")

    def __repr__(self):
        s = DeviceFormat.__repr__(self)
        s += ("  priority = %(priority)s  label = %(label)s" %
              {"priority": self.priority, "label": self.label})
        return s

    @property
    def dict(self):
        d = super(SwapSpace, self).dict
        d.update({"priority": self.priority, "label": self.label})
        return d

    @property
    def formattable(self):
        return super(SwapSpace, self).formattable and self._plugin.available

    @property
    def supported(self):
        return super(SwapSpace, self).supported and self._plugin.available

    @property
    def controllable(self):
        return super(SwapSpace, self).controllable and self._plugin.available

    def labeling(self):
        """Returns True as mkswap can write a label to the swap space."""
        return True

    def relabels(self):
        """Returns True as mkswap can write a label to the swap space."""
        return True and self._plugin.available

    def label_format_ok(self, label):
        """Returns True since no known restrictions on the label."""
        return True

    def write_label(self, dry_run=False):
        """ Create a label for this format.

            :raises: SwapSpaceError

            If self.label is None, this means accept the default, so raise
            an SwapSpaceError in this case.

            Raises a SwapSpaceError if the label can not be set.
        """

        if not self._plugin.available:
            raise SwapSpaceError("application to set label on swap format is not available")

        if not dry_run:
            if not self.exists:
                raise SwapSpaceError("swap has not been created")

            if not os.path.exists(self.device):
                raise SwapSpaceError("device does not exist")

            if self.label is None:
                raise SwapSpaceError("makes no sense to write a label when accepting default label")

            if not self.label_format_ok(self.label):
                raise SwapSpaceError("bad label format")

            blockdev.swap.mkswap(self.device, self.label)

    label = property(lambda s: s._get_label(), lambda s, l: s._set_label(l),
                     doc="the label for this swap space")

    def uuid_format_ok(self, uuid):
        """Check whether the given UUID is correct according to RFC 4122."""
        return fsuuid.FSUUID._check_rfc4122_uuid(uuid)

    def _set_priority(self, priority):
        # pylint: disable=attribute-defined-outside-init
        if priority is None:
            self._priority = -1
            return

        if not isinstance(priority, int) or not -1 <= priority <= 32767:
            # -1 means "unspecified"
            raise ValueError("swap priority must be an integer between -1 and 32767")

        self._priority = priority

    def _get_priority(self):
        return self._priority

    priority = property(_get_priority, _set_priority,
                        doc="The priority of the swap device")

    def _get_options(self):
        opts = ""
        if self.priority is not None and self.priority != -1:
            opts += "pri=%d" % self.priority

        return opts

    def _set_options(self, options):
        if not options:
            self.priority = None
            return

        for option in options.split(","):
            (opt, equals, arg) = option.partition("=")
            if equals and opt == "pri":
                try:
                    self.priority = int(arg)
                except ValueError:
                    log.info("invalid value for swap priority: %s", arg)

    @property
    def status(self):
        """ Device status. """
        return self.exists and blockdev.swap.swapstatus(self.device)

    def _setup(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        blockdev.swap.swapon(self.device, priority=self.priority)

    def _teardown(self, **kwargs):
        """ Close, or tear down, a device. """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        blockdev.swap.swapoff(self.device)

        udev.settle()

    def _create(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        if self.uuid is None:
            blockdev.swap.mkswap(self.device, label=self.label)
        else:
            if not self.uuid_format_ok(self.uuid):
                raise FSWriteUUIDError("bad UUID format for swap filesystem")
            blockdev.swap.mkswap(self.device, label=self.label,
                                 extra={"-U": self.uuid})


register_device_format(SwapSpace)



# i18n.py
# Internationalization functions
#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>

__all__ = ["_", "N_", "P_"]

import gettext
import locale
import six

# Create and cache a translations object for the current LC_MESSAGES value
_cached_translations = {}


def _get_translations():
    # Use setlocale instead of getlocale even though that looks like it makes no sense,
    # since this way we're just reading environment variables instead of mandating some
    # sequence of setlocale calls or whatever, since this is also how the gettext functions
    # behave. This differs from the behavior of gettext.find if $LANGUAGE is being used, but
    # on the other hand no one uses $LANGUAGE.
    lc_messages = locale.setlocale(locale.LC_MESSAGES, None)
    if lc_messages not in _cached_translations:
        _cached_translations[lc_messages] = gettext.translation("blivet", fallback=True)
    return _cached_translations[lc_messages]


N_ = lambda x: x

# In Python 2, return the translated strings as unicode objects.
# yes, pylint, the lambdas are necessary, because I want _get_translations()
# evaluated on every call.
# pylint: disable=unnecessary-lambda,redundant-u-string-prefix
if six.PY2:
    _ = lambda x: _get_translations().ugettext(x) if x != "" else u""
    P_ = lambda x, y, z: _get_translations().ungettext(x, y, z)
else:
    _ = lambda x: _get_translations().gettext(x) if x != "" else ""
    P_ = lambda x, y, z: _get_translations().ngettext(x, y, z)



# fsmount.py
# Filesystem mounting classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import os

from ..errors import FSError
from ..flags import flags
from .. import util
from ..formats import fslib

from . import availability
from . import fstask
from . import task


class FSMount(task.BasicApplication, fstask.FSTask):

    """An abstract class that represents filesystem mounting actions. """
    description = "mount a filesystem"

    options = ["defaults"]
    # type argument to pass to mount, if different from filesystem type
    fstype = None

    ext = availability.MOUNT_APP

    # TASK methods

    @property
    def _has_driver(self):
        """ Is there a filesystem driver in the kernel modules directory. """
        modpath = os.path.realpath(os.path.join("/lib/modules", os.uname()[2]))
        if os.path.isdir(modpath):
            modname = "%s.ko" % self.mount_type
            for _root, _dirs, files in os.walk(modpath):
                if any(x.startswith(modname) for x in files):
                    return True
        return False

    @property
    def _can_mount(self):
        return (self.mount_type in fslib.kernel_filesystems) or \
            (os.access("/sbin/mount.%s" % (self.mount_type,), os.X_OK)) or \
            self._has_driver

    @property
    def _availability_errors(self):
        errors = super(FSMount, self)._availability_errors

        if not self._can_mount:
            errors.append("mounting filesystem %s is not supported" % self.mount_type)
        return errors

    # IMPLEMENTATION methods

    @property
    def mount_type(self):
        """ Mount type string to pass to mount command.

            :returns: mount type string
            :rtype: str
        """
        return self.fstype or self.fs._type

    def _modify_options(self, options):
        """ Any mandatory options can be added in this method.

            :param str options: an option string
            :returns: a modified option string
            :rtype: str
        """
        return options

    def mount_options(self, options):
        """ The options used for mounting.

           :param options: mount options
           :type options: str or NoneType
           :returns: the options used by the task
           :rtype: str
        """
        if not options or not isinstance(options, str):
            options = self.fs.mountopts or ",".join(self.options)

        return self._modify_options(options)

    def do_task(self, mountpoint, options=None):
        """Create the format on the device and label if possible and desired.

           :param str mountpoint: mountpoint that overrides self.mountpoint
           :param options: mount options
           :type options: str or None
        """
        # pylint: disable=arguments-differ
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSError("\n".join(error_msgs))

        try:
            rc = util.mount(self.fs.device, mountpoint,
                            fstype=self.mount_type,
                            options=self.mount_options(options))
        except OSError as e:
            raise FSError("mount failed: %s" % e)

        if rc:
            raise FSError("mount failed: %s" % rc)


class AppleBootstrapFSMount(FSMount):
    fstype = "hfs"


class BindFSMount(FSMount):

    @property
    def _availability_errors(self):
        errors = []
        if not self.ext.available:
            errors.append("application %s is not available" % self.ext)

        return errors

    def _modify_options(self, options):
        return ",".join(["bind", options])


class DevPtsFSMount(FSMount):
    options = ["gid=5", "mode=620"]


class FATFSMount(FSMount):
    options = ["umask=0077", "shortname=winnt"]


class EFIFSMount(FATFSMount):
    fstype = "vfat"


class HFSPlusMount(FSMount):
    fstype = "hfsplus"


class Iso9660FSMount(FSMount):
    options = ["ro"]


class UDFFSMount(FSMount):
    options = ["ro"]


class NoDevFSMount(FSMount):

    @property
    def mount_type(self):
        return self.fs.device


class NFSMount(FSMount):

    @property
    def _availability_errors(self):
        return ["nfs filesystem can't be mounted"]


class NTFSMount(FSMount):
    options = ["default", "ro"]


class SELinuxFSMount(NoDevFSMount):

    @property
    def _availability_errors(self):
        errors = super(SELinuxFSMount, self)._availability_errors
        if not flags.selinux:
            errors.append("selinux not enabled")
        return errors


class StratisXFSMount(FSMount):
    fstype = "xfs"


class TmpFSMount(NoDevFSMount):

    def _modify_options(self, options):
        # This duplicates some code in fs.TmpFS._get_options.
        # There seems to be no way around that.
        if self.fs._accept_default_size:
            size_opt = None
        else:
            size_opt = self.fs._size_option(self.fs._size)
        return ",".join(o for o in (options, size_opt) if o)

    @property
    def _availability_errors(self):
        errors = []
        if not self.ext.available:
            errors.append("application %s is not available" % self.ext)

        return errors
